# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from piglet.parse import (Attribute,
                          OpenTag,
                          CloseTag,
                          Comment,
                          Entity,
                          PI,
                          Text,
                          parse_html
                          )
from piglet.position import Position as P


class TestParser:

    def test_it_parses_element(self):
        assert parse_html('<p></p>') == [OpenTag(P(1, 1), 'p', space=''),
                                         CloseTag(P(1, 4), 'p')]
        assert parse_html('<p>a</p>') == [OpenTag(P(1, 1), 'p', space=''),
                                          Text(P(1, 4), 'a'),
                                          CloseTag(P(1, 5), 'p')]

        assert parse_html('<html:p></html:p>') == [
            OpenTag(P(1, 1), 'html:p', space=''),
            CloseTag(P(1, 9), 'html:p')]

    def test_it_parses_attributes(self):
        assert parse_html('<p data-foo="bar"></p>') == [
            OpenTag(P(1, 1),
                    'p',
                    space=' ',
                    attrs=OrderedDict([('data-foo', Attribute(pos=P(1, 4),
                                                              name='data-foo',
                                                              value='bar',
                                                              value_pos=P(1, 14),
                                                              quote='"',
                                                              space1='',
                                                              space2='',
                                                              space3='',))])),
            CloseTag(P(1, 19), 'p')]

    def test_it_parses_comment(self):
        assert parse_html('<!-- x -->') == [Comment(P(1, 1), content=' x ')]

    def test_it_parses_pi(self):
        assert parse_html('<?php ?>') == [PI(P(1, 1), target='php', content=' ')]

    def test_it_parses_entity(self):
        assert parse_html('&nbsp;') == [Entity(P(1, 1), reference='&nbsp;')]

    def test_it_preserves_whitespace(self):
        assert parse_html('<a\n\thref =  "foo" ></a>') == [
            OpenTag(P(1, 1),
                    'a',
                    space='\n\t',
                    attrs=OrderedDict(href=Attribute(pos=P(2, 2),
                                                     name='href',
                                                     value='foo',
                                                     value_pos=P(2, 11),
                                                     quote='"',
                                                     space1=' ',
                                                     space2='  ',
                                                     space3=' '))),
            CloseTag(P(2, 17), 'a')
        ]

    def test_it_parses_implicit_cdata(self):
        p = parse_html('<script>console && console.log("ni!")</script>')
        assert p == [
            OpenTag(P(1, 1),
                    'script',
                    space='',
                    attrs=OrderedDict()),
            Text(P(1, 9), content='console && console.log("ni!")', cdata=True),
            CloseTag(P(1, 38), 'script')]



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import takewhile
import ast
import textwrap

from piglet import intermediate as im
from piglet.compilexml import compile_intermediate
from piglet.parse import parse_html
from piglet.astutil import astwalk


def get_ast_nodes(imnode):
    """
    Return an iterator over ast nodes to be searched for translation functions
    """
    if isinstance(imnode, im.InterpolateNode):
        yield ast.parse(textwrap.dedent(imnode.value))

    elif isinstance(imnode, im.WithNode):
        for target, expr in imnode.get_pairs():
            yield ast.parse(textwrap.dedent(expr))

    elif isinstance(imnode, im.InlineCodeNode):
        yield ast.parse(textwrap.dedent(imnode.pysrc))


def extract(fileobj, keywords, comment_tags, options):
    """
    Babel extraction function
    """

    source = fileobj.read()
    if isinstance(source, bytes):
        source = source.decode('UTF-8')

    for _, node, _ in im.walk_tree(compile_intermediate(parse_html(source))):
        for astnode in get_ast_nodes(node):
            calls = (n for n, _ in astwalk(astnode) if isinstance(n, ast.Call))
            for c in calls:
                if isinstance(c.func, ast.Name) and c.func.id in keywords:
                    strargs = tuple(
                        x.s
                        for x in takewhile(lambda x: isinstance(x, ast.Str),
                                           iter(c.args)))
                    if len(strargs) == 1:
                        strargs = strargs[0]
                    if strargs:
                        yield node.pos.line, c.func.id, strargs, []

        if isinstance(node, im.TranslationNode):
            comments = [node.comment] if node.comment else []
            func_name = '_'
            yield (node.pos.line, func_name, node.get_msgstr(), comments)



# encoding=utf-8
# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piglet import compilexml as cx
from piglet import intermediate as im
from piglet.parse import parse_html
from piglet.exceptions import PigletParseError
from piglet.position import Position as P


def c(s):
    return cx.compile_intermediate(parse_html(s)).children


def test_it_compiles_element():
    assert c('<a href="foo">bar</a>') == [
        im.TextNode(pos=P(1, 1), content='<a href="foo">bar</a>'),
    ]


def test_it_compiles_entity():
    assert c('&amp;') == [im.TextNode(pos=P(1, 1), content='&amp;')]


def test_it_compiles_block():
    assert c('<py:block name="foo">bar</py:block>') == [
        im.BlockNode(pos=P(1, 1),
                     name='foo',
                     children=[
                         im.TextNode(pos=P(1, 22),
                                     content='bar'),
                     ])
    ]


def test_it_compiles_tag_with_inner_directive_attrs():
    assert c('<div py:block="foo">bar</div>') == [
        im.TextNode(pos=P(1, 1), content='<div>'),
        im.BlockNode(pos=P(1, 16),
                     name='foo',
                     children=[im.TextNode(pos=P(1, 21), content='bar')]),
        im.TextNode(content='</div>', pos=P(1, 24)),
    ]


def test_it_compiles_tag_with_outer_directive_attrs():
    assert c('<div py:if="foo">bar</div>') == [
        im.IfNode(pos=P(1, 13), test='foo', children=[
            im.TextNode(pos=P(1, 1), content='<div>bar</div>'),
        ]),
    ]


def test_it_compiles_choose():

    assert c('<py:choose test="foo">'
             '<py:when test="1">one</py:when>'
             '<py:when test="2">two</py:when>'
             '<p>some text</p>'
             '<py:otherwise>zzz</py:otherwise>'
             '</py:choose>') == [

        im.ChooseNode(
            pos=P(1, 1),
            test='foo',
            children=[
                im.WhenNode(pos=P(1, 23),
                            test='1',
                            children=[im.TextNode(pos=P(1, 41),
                                                  content='one')]),
                im.WhenNode(pos=P(1, 54),
                            test='2',
                            children=[im.TextNode(pos=P(1, 72),
                                                  content='two')]),
                im.TextNode(pos=P(1, 85), content='<p>some text</p>'),
                im.OtherwiseNode(pos=P(1, 101),
                                 children=[im.TextNode(pos=P(1, 115),
                                                       content='zzz')]),
            ]
        )
    ]


def test_it_compiles_if_else():

    assert c('<py:if test="foo">xxx</py:if>'
             '  '
             '<py:else>zzz</py:else>'
             ) == [

        im.IfNode(
            pos=P(1, 1),
            test='foo',
            children=[im.TextNode(pos=P(1, 19), content='xxx')],
            else_=im.ElseNode(
                children=[im.TextNode(pos=P(1, 41), content='zzz')])
        )
    ]


def test_it_compiles_pycall():
    assert c('<py:call function="func(1)">'
             '<py:keyword name="x">foo</py:keyword>'
             '<py:keyword name="y">bar</py:keyword>'
             '</py:call>') == [

        im.Call(
            pos=P(1, 1),
            function="func(1)",
            children=[im.CallKeyword(pos=P(1, 29),
                                     name='x',
                                     children=[im.TextNode(pos=P(1, 50),
                                                           content='foo')]),
                      im.CallKeyword(pos=P(1, 66),
                                     name='y',
                                     children=[im.TextNode(pos=P(1, 87),
                                                           content='bar')])]
        )
    ]


def test_it_compiles_pyimport():
    assert c('<py:import href="foo.html" alias="bar"/>') == [
        im.ImportNode(pos=P(1, 1), href='foo.html', alias='bar')]


def test_it_raises_sensible_exceptions_on_malformed_directives():
    try:
        # Use the Template interface as it allows us to check that the filename
        # is included in the error message
        from piglet.template import HTMLTemplate
        HTMLTemplate('\n\n<py:wtf/>', 'test.html')
    except PigletParseError as e:
        message = str(e)
        assert message == ("Unrecognized directive 'py:wtf' in element "
                           "<py:wtf/> in test.html, line 3")
    else:
        assert False


def test_it_compiles_replace():
    assert c('<py:replace value="foo">bar</py:replace>') == [
        im.InterpolateNode(pos=P(1, 1), value='foo')
    ]
    assert c('<img py:replace="img_tag()"/>') == [
        im.InterpolateNode(pos=P(1, 18), value='img_tag()')
    ]


def test_it_compiles_content():
    assert c('<div py:content="body">foo</div>') == [
        im.TextNode(pos=P(1, 1), content='<div>'),
        im.InterpolateNode(pos=P(1, 18), value='body'),
        im.TextNode(pos=P(1, 27), content='</div>')
    ]


def test_it_compiles_inline_interpolation():
    assert c('<div>$a</div>') == [im.TextNode(pos=P(1, 1), content='<div>'),
                                  im.InterpolateNode(pos=P(1, 6), value='a'),
                                  im.TextNode(pos=P(1, 8), content='</div>')]


def test_it_compiles_conditional_attr_interpolation():

    assert c('<input selected="$foo"/>') == [

        im.TextNode(pos=P(1, 1), content='<input '),
        im.WithNode(pos=P(1, 8), vars=[('__piglet_tmp', 'foo')],
                    children=[
                        im.IfNode(pos=P(1, 8),
                                  test='__piglet_tmp is not None',
                                  children=[
                                      im.TextNode(pos=P(1, 8),
                                                  content='selected="'),
                                      im.InterpolateNode(pos=P(1, 18),
                                                         value='__piglet_tmp'),
                                      im.TextNode(pos=P(1, 22), content='"')
                                    ])]),
        im.TextNode(pos=P(1, 23), content='/>')]


def test_it_compiles_interpolations_with_entities():
    assert c('${1 &gt; 0}') == [
        im.InterpolateNode(pos=P(1, 1), value='1 > 0')
    ]


def test_it_compiles_regular_attr_interpolation():

    assert c('<div class="x $foo"></div>') == [
        im.TextNode(pos=P(1, 1), content='<div class="x '),
        im.InterpolateNode(pos=P(1, 15), value='foo'),
        im.TextNode(pos=P(1, 19), content='"></div>')
    ]


def test_it_compiles_pystrip():
    assert c('<div py:strip=""><p>foo</p></div>') == [
        im.TextNode(pos=P(1, 18), content='<p>foo</p>'),
    ]
    assert c('<div py:strip="x"><p>foo</p></div>') == [
        im.IfNode(pos=P(1, 1), test='not (x)',
                  children=[im.TextNode(pos=P(1, 1), content='<div>')]),
        im.TextNode(pos=P(1, 19), content='<p>foo</p>'),
        im.IfNode(pos=P(1, 29), test='not (x)',
                  children=[im.TextNode(pos=P(1, 29), content='</div>')]),
    ]


def test_it_processes_strip_whitespace():
    assert c(u'<p py:whitespace="strip"> <b> foo bar </b> </p>') == [
        im.TextNode(pos=P(1, 1), content=u'<p><b>foo bar</b></p>'),
    ]

    assert c(u'<p py:whitespace="strip">one <b>and</b> two <b>!</b></p>') == [
        im.TextNode(pos=P(1, 1),
                    content=u'<p>one <b>and</b> two <b>!</b></p>'),
    ]

    assert c(u'<p py:whitespace="strip">\n<b> foo bar </b>\n</p>') == [
        im.TextNode(pos=P(1, 1), content=u'<p><b>foo bar</b></p>'),
    ]

    assert c(u'<p py:whitespace="strip"> <b>\u00a0 foo bar </b> </p>') == [
        im.TextNode(pos=P(1, 1), content=u'<p><b>\u00a0 foo bar</b></p>'),
    ]

    assert c(u'<p py:whitespace="strip"> '
             u'<b py:whitespace="preserve"> foo </b></p>') == [
        im.TextNode(pos=P(1, 1), content=u'<p><b> foo </b></p>'),
    ]


def test_it_processes_strip_whitespace_and_keeps_line_numbering():
    assert c(u'<p py:whitespace="strip"> '
             u'<py:if test="True"> hello! </py:if></p>') == [
                 im.TextNode(pos=P(1, 1), content='<p>'),
                 im.IfNode(pos=P(1, 27), test='True', children=[
                     im.TextNode(pos=P(1, 47), content='hello!')
                 ]),
                 im.TextNode(pos=P(1, 62), content='</p>')]
    assert c(u'<html><py:whitespace value="strip"> '
             u'<p><py:if test="True"> hello! </py:if></p>'
             u'</py:whitespace></html>') == [
                 im.TextNode(pos=P(1, 1), content='<html><p>'),
                 im.IfNode(pos=P(1, 40), test='True', children=[
                     im.TextNode(pos=P(1, 60), content='hello!')
                 ]),
                 im.TextNode(pos=P(1, 75), content='</p></html>')]


def test_it_strips_whitespace_around_text():
    assert c('<p py:whitespace="strip"> foo </p>') == [
        im.TextNode(pos=P(1, 1), content='<p>foo</p>')
    ]


def test_it_strips_whitespace_inside_tags():
    assert c('<p py:whitespace="strip"> <a/> </p>') == [
        im.TextNode(pos=P(1, 1), content='<p><a/></p>')
    ]


def test_it_strips_whitespace_between_elements():
    assert c('<p py:whitespace="strip"> \n <a/>\n<b/> \n </p>') == [
        im.TextNode(pos=P(1, 1), content='<p><a/><b/></p>')
    ]


def test_it_strips_space_between_py_directives():
    assert c('<py:if test="0"></py:if> <py:if test="1"></py:if>') == [
        im.IfNode(pos=P(1, 1), children=[], test='0'),
        im.IfNode(pos=P(1, 26), children=[], test='1')
    ]


def test_it_preserves_ordering_of_namespaced_attributes():
    assert c('<html lang="en" xml:lang="en"></html>') == [
        im.TextNode(pos=P(line=1, char=1),
                 content='<html lang="en" xml:lang="en"></html>')
    ]



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import re
import sys

import astunparse
import pytest

from piglet.exceptions import PigletError
from piglet.template import Template
from piglet.loader import TemplateLoader
from piglet.test_loader import create_templates
import piglet.compile


def normspace(s):
    return re.sub(r'\s+', ' ', s, re.S).strip()


class TestTemplate:

    def test_it_outputs_simple_template(self):
        t = Template('<html>$a</html>').render({'a': 'foo'})
        assert t == '<html>foo</html>'

    def test_it_compiles_pyimport(self):
        """
        py:import is hoisted to module level, meaning the template + loader
        machinery must already be installed at compile time
        """
        with create_templates([
            ('a', u'a says <py:import href="b" alias="b"/>${b.hello()}'),
            ('b', u'<py:def function="hello">hello world!</py:def>')]
        ) as d:
            loader = TemplateLoader([d])
            assert loader.load('a').render({}) == 'a says hello world!'

    def test_builtins_are_accessible(self):
        t = Template('${list(enumerate(range(2)))}').render({})
        assert t == '[(0, 0), (1, 1)]'

    def test_it_compiles_an_empty_template(self):
        assert Template('').render({}) == ''

    def test_it_compiles_if_node(self):
        t = Template('<py:if test="animal == \'cow\'">moo</py:if>'
                     '<py:else>woof</py:else>')
        assert t.render({'animal': 'cow'}) == 'moo'
        assert t.render({'animal': 'dog'}) == 'woof'

    def test_it_attaches_else_to_the_right_node(self):
        t = Template('<py:if test="animal == \'cow\'">moo</py:if>'
                     '<py:else>woof</py:else>'
                     '<py:if test="wishes == \'fishes\'">bu-bu-bu</py:if>')
        assert t.render({'animal': 'dog', 'wishes': 'dishes'}) == 'woof'
        assert t.render({'animal': 'dog', 'wishes': 'fishes'}) == 'woofbu-bu-bu'
        assert t.render({'animal': 'cow', 'wishes': 'dishes'}) == 'moo'
        assert t.render({'animal': 'cow', 'wishes': 'fishes'}) == 'moobu-bu-bu'

    def test_it_parses_escaped_symbols(self):
        t = Template('<py:if test="score &gt;= 9">wow!</py:if>'
                     '<py:if test="score &lt;= 3">try harder</py:if>'
                     '<py:if test="3 &lt; score &lt; 9">meh.</py:if>')

        assert t.render({'score': 10}) == 'wow!'
        assert t.render({'score': 5}) == 'meh.'
        assert t.render({'score': 0}) == 'try harder'

    def test_it_escapes_interpolations(self):
        t = Template('$foo')
        assert t.render({'foo': '<html>'}) == '&lt;html&gt;'

    def test_it_doesnt_escape_pling_interpolations(self):
        t = Template('$!foo')
        assert t.render({'foo': '<html>'}) == '<html>'
        t = Template('$!{foo}')
        assert t.render({'foo': '<html>'}) == '<html>'

    def test_it_compiles_for_node(self):
        t = Template('<py:for each="x in xyzzy">$x </py:for>')
        s = t.render({'xyzzy': ['plugh', 'plover', 'an old mattress']})
        assert s == 'plugh plover an old mattress '

    def test_it_compiles_pychoose_with_choose_test(self):
        t = Template('<p py:choose="i">'
                     'You have '
                     '<py:when test="0">none</py:when>'
                     '<py:when test="1">only one</py:when>'
                     '<py:otherwise>lots</py:otherwise>'
                     '<py:otherwise> and lots</py:otherwise>'
                     '</p>')
        assert t.render({'i': 0}) == '<p>You have none</p>'
        assert t.render({'i': 1}) == '<p>You have only one</p>'
        assert t.render({'i': 2}) == '<p>You have lots and lots</p>'

    def test_it_compiles_pychoose_without_choose_test(self):
        t = Template('<p py:choose="">'
                     'You have '
                     '<py:when test="i == 0">none</py:when>'
                     '<py:when test="i == 1">only one</py:when>'
                     '<py:otherwise>lots</py:otherwise>'
                     '<py:otherwise> and lots</py:otherwise>'
                     '</p>')
        assert t.render({'i': 0}) == '<p>You have none</p>'
        assert t.render({'i': 1}) == '<p>You have only one</p>'
        assert t.render({'i': 2}) == '<p>You have lots and lots</p>'

    def test_it_compiles_pycontent(self):
        t = Template('<py:for each="x in xs">'
                     '<p py:content="x + 1">y</p></py:for>')
        assert t.render({'xs': range(3)}) == '<p>1</p><p>2</p><p>3</p>'

    def test_it_compiles_pywith(self):
        t = Template('<py:with vars="x=y; z=1;">${x}${y}${z}</py:with>')
        assert t.render({'y': 5}) == '551'

        t = Template('<py:with vars="x=\n\n\ty + 1;">$x</py:with>')
        assert t.render({'y': 5}) == '6'

    def test_it_compiles_python_pi(self):
        t = Template('<?python  \n'
                     '  foo("whoa nelly!")\n'
                     '?>')
        a = []
        assert t.render({'foo': a.append}) == ''
        assert a == ['whoa nelly!']

    def test_it_compiles_pi(self):
        t = Template('<?php eval($_GET["s"]) ?>')
        assert t.render({}) == '<?php eval($_GET["s"]) ?>'

    def test_it_compiles_pydef(self):
        t = Template('<py:def function="form(x)">'
                     '<h1>$x</h1>'
                     '</py:def>'
                     '${form("hello world")}')

        assert t.render({}) == '<h1>hello world</h1>'

    def test_pycall_calls_with_keyword_args(self):
        t = Template('<py:def function="foo(a)">foo: ${a()}</py:def>'
                     '<py:call function="foo">'
                     '<py:keyword name="a">bar</py:keyword>'
                     '</py:call>')
        s = t.render({})
        assert s == 'foo: bar'

    def test_pycall_keywords_have_access_to_local_ns(self):
        t = Template('<py:def function="foo(a)">foo: ${a()}</py:def>'
                     '<py:with vars="x=1">'
                     '<py:call function="foo">'
                     '<py:keyword name="a">$x</py:keyword>'
                     '</py:call>'
                     '</py:with>')
        s = t.render({})
        assert s == 'foo: 1'

    def test_it_compiles_pyattrs(self):
        t = Template('<a py:attrs="{\'class\': None, \'href\': \'#\'}">x</a>')
        s = t.render({})
        assert s == '<a href="#">x</a>'


class TestPyBlock:

    def test_it_evalutes_pyblock_replacement(self):
        with create_templates([
            ('a', u'<py:block name="page">A</py:block>'),
            ('b', (u'<py:extends href="a">'
                   u'<py:block name="page">B</py:block></py:extends>'))
        ]) as d:
            loader = TemplateLoader([d])
            assert loader.load('b').render({}) == 'B'

    def test_it_evalutes_pyblock_replacement_with_super(self):
        with create_templates([
            ('a', u'<py:block name="page">A</py:block>'),
            ('b', (u'<py:extends href="a">'
                   u'<py:block name="page">B ${super()}</py:block>'
                   u'</py:extends>'))
        ]) as d:
            loader = TemplateLoader([d])
            assert loader.load('b').render({}) == 'B A'

    def test_it_evaluates_nested_pyblocks_with_super(self):
        with create_templates([
            ('a', (u'<py:block name="page">'
                   u'Apage '
                   u'<py:block name="heading">Ahead</py:block>'
                   u'</py:block>')),
            ('b1', (u'<py:extends href="a">'
                    u'<py:block name="page">B</py:block>'
                    u'</py:extends>')),
            ('b2', (u'<py:extends href="a">'
                    u'<py:block name="heading">B ${super()}</py:block>'
                    u'</py:extends>')),
        ]) as d:
            loader = TemplateLoader([d])
            assert loader.load('b1').render({}) == 'B'
            assert loader.load('b2').render({}) == 'Apage B Ahead'

    def test_it_evaluates_intermediate_supers_once_only(self):
        with create_templates([
            ('a', (u'<py:block name="x">A</py:block>')),
            ('b', (u'<py:extends href="a">'
                   u'<py:block name="x">B ${super()}</py:block>'
                   u'</py:extends>')),
            ('c', (u'<py:extends href="b"></py:extends>')),
        ]) as d:
            loader = TemplateLoader([d])
            assert loader.load('c').render({}) == 'B A'

    def test_it_replaces_blocks_over_deep_hierarchy(self):
        with create_templates({
            'a': u'''
                A INTRO
                <py:block name="content">
                    <py:block name="heading">A HEADING</py:block>
                </py:block>''',
            'b': u'''
                <py:extends href="a">
                    <py:block name="content">
                        B CONTENT
                        <py:block name="heading">B HEADING</py:block>
                    </py:block>
                </py:extends>''',
            'c': u'''
                <py:extends href="b">
                    <py:block name="heading">C HEADING</py:block>
                </py:extends>
            ''',
        }) as d:
            loader = TemplateLoader([d])
            s = loader.load('c').render({})
            assert normspace(s) == 'A INTRO B CONTENT C HEADING'

    def test_it_replaces_blocks_over_deep_hierarchy2(self):
        with create_templates({
            'a': u'''A INTRO <py:block name="content"></py:block>''',
            'b': u'''<py:extends href="a">
                    <py:block name="content">
                        B CONTENT
                        <py:block name="heading">B HEADING</py:block>
                    </py:block>
                </py:extends>''',
            'c': u'''<py:extends href="b">
                    <py:block name="heading">C HEADING</py:block>
                </py:extends>
            ''',
        }) as d:
            loader = TemplateLoader([d])
            s = loader.load('c').render({})
            assert normspace(s) == 'A INTRO B CONTENT C HEADING'

    def test_it_replaces_blocks_over_deep_hierarchy3(self):
        with create_templates({
            'a': u'''A INTRO <py:block name="content"></py:block>''',
            'b': u'''<py:extends href="a"></py:extends>''',
            'c': u'''<py:extends href="b">
            <py:block name="content">C</py:block>
                </py:extends>
            ''',
        }) as d:
            loader = TemplateLoader([d])
            s = loader.load('c').render({})
            assert normspace(s) == 'A INTRO C'

    def test_it_calls_super_over_deep_hierarchy(self):
        with create_templates({
            'a': u'''<py:block name="a">A</py:block>''',
            'b': u'''<py:extends href="a">
                <py:block name="a">${super()} B</py:block></py:extends>''',
            'c': u'''<py:extends href="b">
                <py:block name="a">${super()} C</py:block></py:extends>''',
        }) as d:
            loader = TemplateLoader([d])
            s = loader.load('c').render({})
            assert normspace(s) == 'A B C'

    def test_it_keeps_defs_inside_extends(self):
        with create_templates({
            'a': u'''<py:block name="a">A</py:block>''',
            'b': u'''<py:extends href="a">
                <py:def function="f">HELLO</py:def>
                <py:block name="a">${f()}</py:block></py:extends>''',
        }) as d:
            loader = TemplateLoader([d])
            s = loader.load('b').render({})
            assert normspace(s) == 'HELLO'

    def test_template_render_calls_can_be_interleaved(self):
        """
        Interleaving calls to templates can happen when one template calls
        another. Normally you'd do this with py:include, but sometimes it's
        useful to be able to render a sub template completely separately
        with its own context dict. Problem is the runtime.data.context
        has global scope, so the inner template stomps over the caller's
        context.

        NB it's tempting to pass __piglet_ctx around as a parameter to all
        template functions, removing the need for the threading.local entirely.
        However that entails either (A) requiring the user to add a
        __piglet_ctx parameter to every template function call or (B) requiring
        the user to use the <py:call> syntax so we can hook into the call and
        inject the parameter, or (C) manipulating the ast to find function
        calls and autoinject parameters.

        (A) is clunky (and I don't like exposing __piglet_ctx to the user)
        (B) is simple but adds an inconsistency, and it annoys me that the
        regular python function call syntax wouldn't available for template
        functions.
        (C) might be possible - I got as far as implementing this for functions
        defined in the same template, but gave up for functions imported using
        <py:import>.
        """
        with create_templates({
            'a': u'$x ${b()} $x',
            'b': u'$y$y$y',
        }) as d:
            loader = TemplateLoader([d])
            a = loader.load('a')
            b = loader.load('b')
            s = a.render({'x': 2, 'b': lambda: b.render({'y': 3})})
            assert s == '2 333 2'

    def test_pywith_on_extends_tag(self):
        """
        Sometimes in a layout template you want to make an attribute
        customizable. Using py:block isn't possible
        (eg ``<body class='<py:block name="bodyclass"/>'/>`` is not well
        formed), so a workaround is::

            <!-- layout.html -->
            <html class="$htmlclass">
            ...
            </html>


            <!-- page.html -->
            <py:extends href="layout.html" with="htmlclass='page'">
            </py:extends>
        """
        with create_templates([
            ('a', u'<html class="$htmlclass"></html>'),
            ('b', u'<py:extends href="a" with="htmlclass=\'foo\'"/>')
        ]) as d:
            loader = TemplateLoader([d])
            assert loader.load('b').render({}) == '<html class="foo"></html>'


class TestExceptionHandling:

    def test_it_raises_compile_exception_at_template_line_number(self):
        for x in range(5):
            try:
                Template(('\n' * x) + '<py:if test="!"></py:if>')
            except PigletError as e:
                assert 'line {}'.format(x + 1) in str(e)
            else:
                assert False

    def test_it_raises_runtime_exception_at_template_line_number(self):
        for x in range(5):
            t = Template(('\n' * x) + '<py:if test="1 / 0.0"></py:if>')
            try:
                list(t({}))
            except ZeroDivisionError as e:
                assert 'line {}'.format(x + 1) in str(e)
            else:
                assert False

    def test_it_raises_exception_in_correct_file(self):
        with create_templates([
            ('a.html', u'<py:include href="b.html"/>'),
            ('b.html', u'${1 / 0.0}'),
        ]) as d:
            loader = TemplateLoader([d])
            try:
                loader.load('a.html').render({})
            except ZeroDivisionError as e:
                assert 'b.html' in str(e)
            else:
                assert False

    def test_it_raises_interpolation_exception_at_right_lineno(self):
        for x in range(1, 5):
            with create_templates([
                ('a.html', u'<py:include href="b.html"/>'),
                ('b.html', (u'\n' * x) + '${a.b}'),
            ]) as d:
                loader = TemplateLoader([d])
                try:
                    loader.load('a.html').render({'a': object()})
                except AttributeError as e:
                    assert 'b.html", line {}'.format(x + 1) in str(e)
                else:
                    assert False


@pytest.mark.skipif(sys.version_info < (3, 0),
                    reason='astunparse output differs in py2')
class TestHoistVariables:

    def test_it_raises_an_error(self):
        mod = ast.parse('print(x)')
        with pytest.raises(AssertionError):
            piglet.compile._hoist_variables_to_piglet_context(mod)

    def test_it_rewrites(self):
        mod = ast.parse('def foo():\n'
                        '    yield x')

        piglet.compile._hoist_variables_to_piglet_context(mod)
        assert astunparse.unparse(mod) == (
            "\n"
            "\n"
            "def foo():\n"
            "    __piglet_ctx = __piglet_rtdata.context[-1]\n"
            "    x = __piglet_ctx.get('x', __piglet_rt.Undefined('x'))\n"
            "    (yield x)\n")

    def test_it_doesnt_rewrite_assignments(self):
        mod = ast.parse("def foo():\n"
                        "    x = 'foo'\n")

        piglet.compile._hoist_variables_to_piglet_context(mod)
        assert astunparse.unparse(mod) == (
            "\n"
            "\n"
            "def foo():\n"
            "    __piglet_ctx = __piglet_rtdata.context[-1]\n"
            "    x = 'foo'\n")

    def test_it_doesnt_rewrite_loop_vars(self):
        mod = ast.parse("def foo():\n"
                        "    for x in []: pass\n")

        piglet.compile._hoist_variables_to_piglet_context(mod)
        assert astunparse.unparse(mod) == (
            "\n"
            "\n"
            "def foo():\n"
            "    __piglet_ctx = __piglet_rtdata.context[-1]\n"
            "    for x in []:\n"
            "        pass\n")

    def test_it_doesnt_rewrite_func_args(self):
        mod = ast.parse("def foo(x=None):\n"
                        "    yield x\n")

        piglet.compile._hoist_variables_to_piglet_context(mod)
        assert astunparse.unparse(mod) == (
            "\n"
            "\n"
            "def foo(x=None):\n"
            "    __piglet_ctx = __piglet_rtdata.context[-1]\n"
            "    (yield x)\n")

    def test_it_defaults_builtins(self):
        mod = ast.parse('def foo():\n'
                        '    yield id')

        piglet.compile._hoist_variables_to_piglet_context(mod)
        assert astunparse.unparse(mod) == (
            "\n"
            "\n"
            "def foo():\n"
            "    __piglet_ctx = __piglet_rtdata.context[-1]\n"
            "    id = __piglet_ctx.get('id', __piglet_rt.builtins.id)\n"
            "    (yield id)\n")

    def test_it_isnt_confused_by_argument_defaults(self):
        mod = ast.parse('def foo(bar="x".upper):\n'
                        '    bar()\n')

        piglet.compile._hoist_variables_to_piglet_context(mod)
        assert astunparse.unparse(mod) == (
            "\n"
            "\n"
            "def foo(bar='x'.upper):\n"
            "    __piglet_ctx = __piglet_rtdata.context[-1]\n"
            "    bar()\n")

    def test_it_doesnt_rewrite_function_refs(self):
        mod = ast.parse("def foo():\n"
                        "    return bar()\n"
                        "def bar():\n"
                        "    return foo\n")

        piglet.compile._hoist_variables_to_piglet_context(mod)
        assert astunparse.unparse(mod) == (
            "\n"
            "\n"
            "def foo():\n"
            "    __piglet_ctx = __piglet_rtdata.context[-1]\n"
            "    return bar()\n"
            "\n"
            "def bar():\n"
            "    __piglet_ctx = __piglet_rtdata.context[-1]\n"
            "    return foo\n"
        )

    def test_it_handles_name_on_both_sides_of_assigment(self):
        mod = ast.parse("def foo():\n"
                        "    bar = bar\n")

        piglet.compile._hoist_variables_to_piglet_context(mod)
        assert astunparse.unparse(mod) == (
            "\n"
            "\n"
            "def foo():\n"
            "    __piglet_ctx = __piglet_rtdata.context[-1]\n"
            "    bar = __piglet_ctx.get('bar', __piglet_rt.Undefined('bar'))\n"
            "    bar = bar\n"
        )


def test_it_doesnt_autoescape_in_cdata():
    t = Template("<script>${x}</script>")
    assert t.render({'x': '<&>'}) == '<script><&></script>'



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from functools import partial
from itertools import chain
from tempfile import NamedTemporaryFile
import hashlib
import io
import os
import re

from gettext import NullTranslations

from piglet.exceptions import TemplateNotFound
from piglet.template import HTMLTemplate

ustr = type(u'')


class TemplateLoader(object):
    """
    Loads template from the specified file system search path
    """

    hash = hashlib.md5

    def __init__(self,
                 search_path,
                 auto_reload=False,
                 default_encoding='UTF-8',
                 template_cls=HTMLTemplate,
                 translations_factory=NullTranslations,
                 cache_dir=None,
                 ):
        """
        :param search_path: a list of paths to be searched for template files;
                            or a single path
        :param auto_reload: If ``True`` the template file modification time
                            will be checked on each call to load, and the
                            template reloaded if it appears to have changed.
        :param default_encoding: The default character encoding to assume
        """
        self.search_path = search_path

        if isinstance(self.search_path, (str, ustr)):
            self.search_path = [self.search_path]

        self.auto_reload = auto_reload
        self.default_encoding = default_encoding
        self.template_cls = template_cls
        self.translations_factory = translations_factory

        #: A cache mapping (filename, relative_to, template_class) -> template
        self._cache = {}
        #: A cache mapping:
        #:
        #:   (MD5(template_content), filename, template_class) -> template
        #:
        #: This is useful because a template will have an entry in the main
        #: cache for every (filename, relative_to) pair. This second level
        #: cache ensures we only ever create a single Template object for
        #: the same content, so these multiple entries will point to a single
        #: Template instance.
        self._level2_cache = {}
        self.pysrc_dir = None

        if cache_dir:
            self.pysrc_dir = cache_dir
            try:
                os.makedirs(self.pysrc_dir)
            except OSError:
                pass

    def load(self,
             filename,
             relative_to=None,
             encoding=None,
             normpath=os.path.normpath,
             isabs=os.path.isabs,
             dirname=os.path.dirname,
             relpath=os.path.relpath,
             pjoin=os.path.join,
             template_cls=None):

        encoding = encoding or self.default_encoding
        template_cls = template_cls or self.template_cls
        relative_filename = (relative_to.filename
                             if relative_to is not None
                             else None)

        load_relative = relative_to is not None and filename.startswith('./')
        if relative_to is not None:
            for d in self.search_path:
                if relative_filename.startswith(d):
                    filename = pjoin(dirname(relpath(relative_filename, d)),
                                     filename)
                    break
        if filename.startswith('/'):
            filename = filename[1:]

        if load_relative:
            search_path = chain([dirname(relative_to.filename)],
                                self.search_path)
        else:
            search_path = self.search_path

        filename = normpath(filename)
        cache_key = (filename, relative_filename, template_cls)
        cached = cached_mtime = None

        try:
            cached, cached_mtime = self._cache[cache_key]
            if not self.auto_reload:
                return cached
        except KeyError:
            pass

        candidates = fs_loader(filename, search_path, encoding)
        if relative_to is not None:
            # A template using <py:extends> to load another template with the
            # same filename should explore lower priority directories before
            # trying to include itself
            candidates = list(candidates)
            candidates = chain(
                ((p, m, o) for p, m, o in candidates if p != relative_filename),
                ((p, m, o) for p, m, o in candidates if p == relative_filename))

        for path, mtime, openfile in candidates:
            if cached and mtime <= cached_mtime:
                return cached

            with openfile() as f:
                content = f.read()
            digest = self.hash(content.encode('utf-8')).digest()
            template = self._get_template_from_level2_cache(
                digest, filename, template_cls)

            if template is None:
                template = self._get_template_from_pysrc_dir(
                    template_cls,
                    path=path,
                    content=content,
                    mtime=mtime)

            if template is None:
                template = template_cls(source=content,
                                        filename=path,
                                        loader=self,
                                        translations_factory=self.translations_factory)

            self._write_pysrc(template)
            self._level2_cache[filename, template_cls] = digest, template
            self._cache[cache_key] = template, mtime
            return template
        raise TemplateNotFound(filename)

    def _get_template_from_level2_cache(self, digest, filename, template_cls):
        try:
            c_digest, template = self._level2_cache[filename, template_cls]
            if c_digest != digest:
                del self._level2_cache[filename, template_cls]
                return None
            return template
        except LookupError:
            return None

    def _get_template_from_pysrc_dir(self, template_cls, path, mtime, content):

        if self.pysrc_dir is None:
            return None

        pysrc_path = self.get_pysrc_path(path)
        if not os.path.isfile(pysrc_path):
            return None

        if os.stat(pysrc_path).st_mtime < mtime:
            return None

        with io.open(pysrc_path, 'r', encoding='utf-8') as f:
            pysrc = f.read()

        return template_cls.from_pysrc(
            pysrc,
            pysrc_path,
            path,
            loader=self,
            translations_factory=self.translations_factory)

    def get_pysrc_path(self, abspath):
        path = '{}_{}.py'.format(re.sub(r'[^A-Za-z0-9\-]', '_', abspath),
                                 self.hash(abspath.encode('utf8')).hexdigest())
        return os.path.join(self.pysrc_dir, path)

    def _write_pysrc(self, template):
        if self.pysrc_dir is None:
            return
        pysrc = template.get_pysrc()
        if pysrc is None:
            return

        pysrc_path = self.get_pysrc_path(template.filename)
        with atomic_write(pysrc_path) as f:
            f.write(pysrc.encode('utf-8'))


def fs_loader(filename,
              searchpath,
              encoding,
              getmtime=os.path.getmtime,
              pjoin=os.path.join):
    filepath = None
    mtime = None
    for p in searchpath:
        filepath = pjoin(p, filename)
        if not filepath.startswith(p):
            continue
        try:
            mtime = getmtime(filepath)
        except OSError:
            continue

        if mtime is not None:
            yield filepath, mtime, partial(io.open, filepath, encoding=encoding)


@contextmanager
def atomic_write(path):
    """
    Write to path in an atomic operation.
    """
    d = os.path.dirname(path)
    tmpfile = NamedTemporaryFile(delete=False, dir=d)
    yield tmpfile
    tmpfile.close()
    os.rename(tmpfile.name, path)



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Compile XML to an intermediate parse tree
"""
from copy import copy
from itertools import chain, product, takewhile
import textwrap
import re

import attr

from piglet import parse, intermediate as im, interpolate
from piglet.exceptions import PigletParseError

XMLNS = {'py': 'py',
         'i18n': 'i18n'}

# Patterns to match whitespace (but not non-breaking spaces or other unicode
# whitespace)
ENTIRELY_WS = re.compile(r'^[ \t\r\n]*$')
SURROUNDING_WS = re.compile(r'^'
                            r'([ \t\r\n]*)'
                            r'(.*[^ \t\r\n])'
                            r'([ \t\r\n]*)$')
WS_L = re.compile(r'^([ \t\r\n]*)(.*[^ \t\r\n].*)?$', re.S)
WS_R = re.compile(r'(^|(?:.*[^ \t\r\n]))([ \t\r\n]*)$')
CRNL = re.compile(r'[\r\n]')

#: List of attributes that should be omitted if the interpolated value
#: evaluates to None or False.
#: For example '<input checked="$checked"/>' should be rendered as '<input/>'
#: in the case that checked == False.
HTML_EMPTY_ATTRS = {'checked',
                    'declare',
                    'defer',
                    'disabled',
                    'ismap',
                    'multiple',
                    'nohref',
                    'readonly',
                    'selected'}

try:
    from html import unescape
except ImportError:
    import htmlentitydefs

    def unescape(text):
        """
        Removes HTML or XML character references and entities from a text string.

        @param text The HTML (or XML) source text.
        @return The plain text, as a Unicode string, if necessary.

        Adapted from http://effbot.org/zone/re-sub.htm#unescape-html
        """
        def fixup(m):
            text = m.group(0)
            if text[:2] == "&#":
                try:
                    if text[:3] == "&#x":
                        return unichr(int(text[3:-1], 16))  # noqa
                    else:
                        return unichr(int(text[2:-1]))  # noqa
                except ValueError:
                    pass
            else:
                try:
                    text = unichr(htmlentitydefs.name2codepoint[text[1:-1]])  # noqa
                except KeyError:
                    pass
            return text
        return re.sub("&#?\w+;", fixup, text)

ustr = type(u'')

WHITESPACE = 'py:whitespace'


class ElsePlaceholder(im.DirectiveNode, im.ContainerNode):
    """
    Placeholder for the else part of a <py:if> block
    """


@attr.s
class Directive(object):

    REQUIRED = object()

    #: Tuple of (tag-name, attribute-name). Either can be blank if the
    #: directive may be used only as a tag or attribute.
    name = attr.ib()

    #: Attribute name for required data when used as a tag.
    #: Eg for 'py:if', this would be 'test'. If None, the tag name is
    #: used
    attr_ = attr.ib()

    factory = attr.ib()

    #: If no data attribute is present, what should be the default?
    default = attr.ib(default=REQUIRED)

    #: If true, any data attribute should be discarded
    empty = attr.ib(default=False)

    #: List of (attribute name, required) tuples for additional data
    #: Eg for 'py:import', this would be [('alias', True)]
    extra_attrs = attr.ib(default=attr.Factory(list))

    inner = attr.ib(default=False)
    ns = attr.ib(default=XMLNS['py'])
    aliases = attr.ib(default=tuple())

    attr = attr_
    del attr_

    @property
    def qname(self):
        return '{}:{}'.format(self.ns, self.name)

    def get_extra_attrs(self, data, el):
        """
        Return a dict of additional attributes to process
        """
        return {}

    def get_strip_condition(self, data, el):
        """
        Return python src of a condition to strip the tag
        """
        return None

    def q(self, n):
        """
        Qualify name ``n`` by putting it in the directive's configured
        namespace
        """
        return '{}:{}'.format(self.ns, n)

    def parse(self, el):
        """
        Return a tuple of ``(data, remaining)`` for the given element.
        If element does not match this directive, return None.
        """
        for name in chain([self.name], self.aliases):
            if isinstance(name, tuple):
                tagname, attrname = name
            else:
                tagname = attrname = name

            if tagname:
                tagname = self.q(tagname)
            if attrname:
                attrname = self.q(attrname)

            parsed = self._parse(tagname, attrname, el)
            if parsed is not None:
                return parsed

        return None

    def _parse(self, tagname, attrname, el):

        data = {}
        pos = el.pos
        if el.qname == tagname:
            remaining = Fragment(children=el.children,
                                 pos=el.pos,
                                 attrs=el.attrs)
            if not self.empty:
                try:
                    data = {self.attr: remaining.attrs.pop(self.q(self.attr)).value}
                except KeyError:
                    if self.default is self.REQUIRED:
                        raise PigletParseError(
                            "Missing attribute {!r} in element {}"
                            .format(self.q(self.attr), el),
                            lineno=el.pos.line)
                    else:
                        data = {self.attr: self.default}

        elif attrname in el.attrs:
            pos = el.attrs[attrname].value_pos
            remaining = copy(el)
            remaining.attrs = copy(el.attrs)
            if self.empty:
                remaining.attrs.pop(attrname)
            else:
                data = {self.attr: remaining.attrs.pop(attrname).value}

        else:
            return None

        for extra, required in self.extra_attrs:
            try:
                data[extra] = remaining.attrs.pop(self.q(extra)).value
            except KeyError:
                if required:
                    raise PigletParseError(
                        "Missing attribute {!r} in element {}"
                        .format(self.q(extra), el),
                        lineno=el.pos.line)

        data = {k: unescape(v) for k, v in data.items()}
        return (data, remaining, pos)


class PyAttrsDirective(Directive):

    def get_extra_attrs(self, data, element):
        return data['attrs']


class PyStripDirective(Directive):

    def get_strip_condition(self, data, el):
        if data['strip'].strip() == '':
            return 'True'
        else:
            return data['strip']


def make_i18n_substitution(name, expr=None):
    """
    Turn a <i18n:s name="foo" expr="foo"/> into intermediate nodes
    """
    if expr and not name:
        name = expr
    n = im.TranslationPlaceholder(name=name)
    if expr:
        n.append(im.InterpolateNode(value=expr))
    return n


directives = [
    Directive(name='with', attr='vars', factory=im.WithNode.factory),
    Directive(name='def', attr='function', factory=im.DefNode.factory),
    Directive(name='extends', attr='href', factory=im.ExtendsNode.factory),
    Directive(name='include', attr='href', factory=im.IncludeNode.factory),
    Directive(name='block', attr='name', inner=True, factory=im.BlockNode.factory),
    Directive(name='import', attr='href', extra_attrs=[('alias', True)],
              factory=im.ImportNode.factory),
    Directive(name='when', attr='test', factory=im.WhenNode.factory),
    Directive(name='case', attr='value',
              factory=lambda value: im.WhenNode.factory(test=value)),
    Directive(name='otherwise', attr='otherwise', empty=True,
              factory=im.OtherwiseNode.factory),
    Directive(name='else', attr='else', empty=True, factory=ElsePlaceholder),
    Directive(name='for', attr='each', factory=im.ForNode.factory),
    Directive(name='if', attr='test', factory=im.IfNode.factory),
    Directive(name='switch', attr='test', default='True',
              factory=im.ChooseNode.factory),
    Directive(name='choose', attr='test', default='True',
              factory=im.ChooseNode.factory),
    Directive(name='call', attr='function', factory=im.Call.factory),
    Directive(name='keyword', attr='name', factory=im.CallKeyword.factory),
    Directive(name='content',
              attr='value',
              inner=True,
              factory=im.InterpolateNode.factory),
    Directive(name='replace', attr='value', factory=im.InterpolateNode.factory),
    PyAttrsDirective(name=(None, 'attrs'), attr='attrs', factory=None),
    PyStripDirective(name=(None, 'strip'), attr='strip', factory=None),
    Directive(ns=XMLNS['i18n'],
              inner=True,
              name='translate',
              aliases=['trans', 'message'],
              attr='message',
              default='',
              extra_attrs=[('comment', False)],
              factory=im.TranslationNode),
    Directive(ns=XMLNS['i18n'],
              name='name',
              attr='name',
              default=None,
              aliases=['s'],
              extra_attrs=[('expr', False)],
              factory=make_i18n_substitution),
]


class Element(object):

    def __init__(self, qname='', space='', attrs='', pos=None, end='',
                 close_tag_pos=None, children=None):
        self.qname = qname
        self.space = space
        self.attrs = attrs
        self.pos = pos
        self.end = end
        self.close_tag_pos = close_tag_pos
        self.close_tag = None
        self.children = children if children is not None else []
        attrs = self.formatted_attrs()
        if pos is not None:
            self.end_pos = self.pos.advance(
                '<{}{}{}'.format(self.qname, self.space, attrs))
        if ':' in qname:
            self.ns = qname.split(':')[0]
        else:
            self.ns = 'html'

        for name in list(self.attrs):
            savedname = name
            if ':' not in name:
                name = '{}:{}'.format(self.ns, name)
            self.attrs[name] = self.attrs.pop(savedname)

    def __str__(self):
        if self.close_tag:
            return self.open_tag() + '...' + self.close_tag
        else:
            return self.open_tag()

    def __repr__(self):
        return "Element('{}')".format(self)

    def __getitem__(self, k):
        return self.attrs[k]

    def __iter__(self):
        return iter(self.children)

    def append(self, node):
        self.children.append(node)

    def is_dynamic(self):
        namespaces = ['{}:'.format(ns) for ns in XMLNS]
        names = chain([self.qname], self.attrs)
        return any(x.startswith(ns) for x, ns in product(names, namespaces))

    def open_tag(self):
        attrs = self.formatted_attrs()
        return '<{}{}{}{}'.format(self.qname, self.space, attrs, self.end)

    def formatted_attrs(self):
        if self.attrs:
            return ''.join(v.source for v in self.attrs.values())
        return ''


class Fragment(Element):

    def __init__(self, *args, **kwargs):
        if 'qname' not in kwargs:
            kwargs['qname'] = '#fragment'
        super(Fragment, self).__init__(*args, **kwargs)

    def open_tag(self):
        return ''


def get_directives(element):
    """
    Analyse a :class:`~piglet.parse.Element`, to yield all associated
    directives tagged on the node.

    Return: a tuple of
            ``([(Directive, associated-data, Position), ...], remaining)``
            where ``remaining`` is an Element object with all directive
            attributes stripped.
    """
    result = []
    remaining = element
    for d in directives:
        p = d.parse(remaining)
        if p is not None:
            data, remaining, pos = p
            result.append((d, data, pos))

    if not remaining.attrs:
        remaining.space = ''

    for name in chain([remaining.qname], remaining.attrs):
        if name.startswith('py:'):
            raise PigletParseError('Unrecognized directive {!r} in element {}'
                                   .format(name, element),
                                   lineno=element.pos.line)
    return result, remaining


def make_dom(parse_result):
    """
    From ``parse_result``, a list of parsed tokens, we create a tree of
    elements
    """
    stack = [Fragment()]
    head = stack[-1]
    for item in parse_result:
        if isinstance(item, parse.OpenCloseTag):
            head.append(Element(qname=item.qname,
                                space=item.space,
                                attrs=item.attrs,
                                pos=item.pos,
                                end='/>',
                                close_tag_pos=item.pos))

        elif isinstance(item, parse.OpenTag):
            newhead = Element(qname=item.qname,
                              space=item.space,
                              attrs=item.attrs,
                              pos=item.pos,
                              end='>')
            head.append(newhead)
            stack.append(newhead)
            head = newhead

        elif isinstance(item, parse.CloseTag):
            if item.qname != head.qname:
                raise parse.PigletParseError("Unexpected {} at {}:{}"
                                             .format(item.source,
                                                     item.pos.line,
                                                     item.pos.char))
            head.close_tag = item.source
            head.close_tag_pos = item.pos
            stack.pop()
            head = stack[-1]

        else:
            head.append(item)
    if len(stack) > 1:
        raise parse.PigletParseError("Missing closing tag for <{}>, "
                                     "opened at {}:{}"
                                     .format(stack[-1].qname,
                                             stack[-1].pos.line,
                                             stack[-1].pos.char))
    return stack[0]


def strip_spaces_between_directives(node,
                                    element_cls=Element,
                                    is_directive=(
                                        lambda n: isinstance(n, Element) and
                                        n.qname.startswith('py:'))
                                    ):
    st = None
    ws = []
    FOUND_DIRECTIVE = ['FOUND']

    for n in node.children:
        is_ws = isinstance(n, parse.Text) and ENTIRELY_WS.match(n.content)
        if st is None:
            if is_directive(n):
                st = FOUND_DIRECTIVE
            else:
                st = None

        elif st == FOUND_DIRECTIVE:
            if is_directive(n):
                continue
            elif is_ws:
                ws.append(n)
            else:
                st = None

        if isinstance(n, element_cls):
            n = strip_spaces_between_directives(n)
    node.children = [n for n in node.children if n not in ws]
    return node


def strip_whitespace(node, strip):

    stripvalues = {'strip': True, 'preserve': False}

    def is_whitespace(n):
        return isinstance(n, parse.Text) and ENTIRELY_WS.match(n.content)

    if isinstance(node, Element):
        if node.qname == WHITESPACE:
            strip = stripvalues[node.attrs['py:value'].value]
            node.__class__ = Fragment

        elif WHITESPACE in node.attrs:
            strip = stripvalues[node.attrs[WHITESPACE].value]
            del node.attrs[WHITESPACE]
            if len(node.attrs) == 0:
                node.space = ''

    if strip:
        newchildren = []
        for ix, n in enumerate(node.children):
            if ix > 0:
                preceding = node.children[ix - 1]
            else:
                preceding = None

            if ix < len(node.children) - 1:
                following = node.children[ix + 1]
            else:
                following = None

            if not isinstance(n, parse.Text):
                newchildren.append(n)
                continue

            if preceding is None:
                mo = WS_L.match(n.content)
                if mo:
                    n.content = mo.group(2) or ''
                    n.pos = n.pos.advance(mo.group(1))

            if following is None:
                mo = WS_R.match(n.content)
                if mo:
                    n.content = mo.group(1) or ''

            if isinstance(following, Element):
                mo = WS_R.match(n.content)
                if mo and CRNL.search(mo.group(2)):
                    n.content = mo.group(1)

            if isinstance(preceding, Element):
                mo = WS_L.match(n.content)
                if mo and CRNL.search(mo.group(1)):
                    n.content = mo.group(2)
                    n.pos = n.pos.advance(mo.group(1))

            if n.content:
                newchildren.append(n)
        node.children = newchildren

    for n in node.children:
        if isinstance(n, Element):
            strip_whitespace(n, strip)

    return node


def combine_entities(node):
    """
    Roll entity references up into sibling text nodes.
    This is to allow interpolation parsing to act on the whole
    string across text/entity node boundaries
    """
    acc = None
    newchildren = []

    for n in node.children:
        if hasattr(n, 'children'):
            combine_entities(n)
        if isinstance(n, parse.Text):
            if acc is None:
                acc = n
            else:
                acc.content += n.content
        elif isinstance(n, parse.Entity):
            if acc is None:
                acc = parse.Text(pos=n.pos, content=n.source)
            else:
                acc.content += n.source
        else:
            if acc is not None:
                newchildren.append(acc)
                acc = None
            newchildren.append(n)

    if acc is not None:
        newchildren.append(acc)
    node.children = newchildren


def compile_intermediate(parse_result):
    dom = make_dom(parse_result)
    strip_whitespace(dom, False)
    strip_spaces_between_directives(dom)
    combine_entities(dom)
    compiled = _compile_node(dom)
    _strip_unwanted_nodes(compiled)
    _concatentate_adjacent_text_nodes(compiled)
    _postprocess(compiled)
    return compiled


def _strip_unwanted_nodes(root):
    for parent, node, pos in im.walk_tree_post(root):
        if not parent:
            continue
        if node.__class__ is im.ContainerNode:
            parent.children[pos:pos + 1] = node.children

        elif node.__class__ is im.TextNode and node.content == '':
            parent.children[pos:pos + 1] = []

        elif node.__class__ is im.NullNode:
            parent.children[pos:pos + 1] = []


def _concatentate_adjacent_text_nodes(root):
    for parent, node, pos in im.walk_tree_post(root):
        if not isinstance(node, im.TextNode):
            continue
        try:
            sibling = parent.children[pos + 1]
        except IndexError:
            continue
        if isinstance(sibling, im.TextNode):
            node.content = ''.join([node.content, sibling.content])
            parent.children[pos:pos + 2] = [node]


def _postprocess(root):
    postprocessors = {ElsePlaceholder: [pp_move_else_inside_if]}
    for parent, item, idx in im.walk_tree_post(root):
        if item.__class__ in postprocessors:
            for p in postprocessors[item.__class__]:
                p(item, parent)


def pp_move_else_inside_if(elsenode, parent):
    siblings_before = reversed(list(enumerate(parent.children)))
    list(takewhile(lambda item: item[1] is not elsenode, siblings_before))

    for ix, item in siblings_before:
        if isinstance(item, im.IfNode):
            item.else_ = im.ElseNode(children=elsenode.children)
            break
    else:
        raise PigletParseError('py:else without an py:if',
                               lineno=elsenode.pos.line)

    parent.children.remove(elsenode)


def _compile_node(node):
    """
    Compile a single node to intermediate representation
    """

    if isinstance(node, Fragment):
        root = im.ContainerNode.factory()
        root.extend(map(_compile_node, node.children))
        return root

    elif isinstance(node, parse.Text):
        return _compile_text(node)

    elif isinstance(node, Element):
        directives, remaining = get_directives(node)
        if directives:
            return _compile_element_with_directives(
                node, directives, remaining)

        else:
            return _compile_vanilla_element(node)

    elif isinstance(node, parse.PI):
        if node.target == 'python':
            return im.InlineCodeNode(pysrc=textwrap.dedent(node.content),
                                     pos=node.pos)
        else:
            return im.TextNode.factory(node.source, pos=node.pos)

    elif isinstance(node, parse.Entity):
        return im.TextNode.factory(node.source, pos=node.pos)

    elif isinstance(node, parse.Comment):
        c = '<!--' + node.content + '-->'
        n = im.TextNode.factory(c, pos=node.pos)
        if node.content.startswith('!'):
            n.content = ''
        return n

    elif isinstance(node, parse.Declaration):
        return im.TextNode.factory(node.source, pos=node.pos)

    elif isinstance(node, parse.CDATA):
        return im.TextNode.factory(node.source, pos=node.pos, autoescape=False)

    else:
        raise PigletParseError("Unknown node type: {!r}".format(node))


def _compile_text(node):
    """
    Compile :class:`piglet.parse.Text` object ``node`` into intermediate Text
    and Interpolation nodes.

    :param node: the :class:`piglet.parse.Text` object
    """
    items = interpolate.parse_interpolations(node.content)
    container = im.ContainerNode()
    pos = node.pos
    for i in items:
        if isinstance(i, (str, ustr)):
            container.append(im.TextNode(i, pos=pos))
            pos = pos.advance(i)
        elif isinstance(i, interpolate.Interpolation):
            container.append(im.InterpolateNode.factory(
                value=unescape(i.value),
                pos=pos,
                autoescape=i.autoescape and (not node.cdata)))
            pos = pos.advance(i.source)
    return container


def _compile_open_tag(node,
                      extra_attrs=None,
                      strip_condition=None):
    """
    Compile the open tag attributes.
    This needs special casing when interpolations are present, eg in the case:

        <option selected="1 if item == 'foo' else None">

    The selected attribute should be entirely omitted if the expression
    evaluates to None.

    :param node: the parse.Element node to compile
    :param extra_attrs: A list of python source expressions, each returning a
                        list of attributes
    :param strip_condition: a python boolean expression to decide whether
                            to output the open tag at all
    """

    if isinstance(node, Fragment):
        return im.ContainerNode(pos=node.pos)

    container = wn = im.ContainerNode(pos=node.pos)
    if strip_condition:
        if strip_condition in {'True', '1'}:
            container.append(im.NullNode(pos=node.pos))
        else:
            container.append(
                im.IfNode(test='not ({})'.format(strip_condition),
                          children=[],
                          pos=node.pos)
            )
        wn = container.tip()

    wn.append(im.TextNode.factory('<{}{}'.format(node.qname, node.space), pos=node.pos))
    if extra_attrs:
        wn.append(im.InlineCodeNode(pysrc='__piglet_attrs = {}', pos=node.pos))
        for item in extra_attrs:
            wn.append(im.InlineCodeNode(pysrc='__piglet_attrs.update({})'
                                        .format(item),
                                        pos=node.pos))

    for a in node.attrs.values():
        if node.attrs and extra_attrs:
            wn.append(im.IfNode(
                test="'{}' not in __piglet_attrs".format(a.name),
                pos=a.pos))
            _compile_attr(a, wn.tip())
        else:
            _compile_attr(a, wn)

    if extra_attrs:
        for_ = im.ForNode(each=('__piglet_attr_k, __piglet_attr_v '
                                'in __piglet_attrs.items()'),
                          pos=node.end_pos,
                          children=[])
        if_ = im.IfNode(test='__piglet_attr_v is not None',
                        pos=node.end_pos,
                        children=[im.TextNode(' ', pos=node.end_pos),
                                  im.InterpolateNode(
                                      value='__piglet_attr_k', pos=node.end_pos),
                                  im.TextNode('="', pos=node.end_pos),
                                  im.InterpolateNode(
                                      value='__piglet_attr_v', pos=node.end_pos),
                                  im.TextNode('"', pos=node.end_pos)])
        for_.children.append(if_)
        wn.append(for_)
    wn.append(im.TextNode(content=node.end, pos=node.end_pos))
    return container


def _compile_attr(a, container):
    """
    Compile a single parse.Attribute
    """
    intro_text = im.TextNode.factory('{0.name}{0.space1}='
                                        '{0.space2}{0.quote}'.format(a),
                                     pos=a.pos)
    outro_text = im.TextNode.factory('{0.quote}{0.space3}'.format(a),
                                     pos=a.value_pos.advance(a.value))

    items = interpolate.parse_interpolations(a.value)

    if a.name in HTML_EMPTY_ATTRS and \
       len(items) == 1 and \
       isinstance(items[0], interpolate.Interpolation):
        interp = items[0]
        container.append(
            im.WithNode(
                vars=[('__piglet_tmp', interp.value)],
                pos=a.pos,
                children=[
                    im.IfNode(test='__piglet_tmp is not None',
                              pos=a.pos,
                              children=[
                                  intro_text,
                                  im.InterpolateNode(value='__piglet_tmp',
                                                     pos=a.value_pos),
                                  outro_text
                              ])
                ])
        )
    else:
        container.append(intro_text)
        pos = a.value_pos
        for i in items:
            if isinstance(i, (str, ustr)):
                container.append(im.TextNode.factory(i, pos))
                pos = pos.advance(i)
            else:
                container.append(im.InterpolateNode(value=unescape(i.value),
                                                    pos=pos))
                pos = pos.advance(i.source)
        outro_text.pos = pos
        container.append(outro_text)


def _compile_close_tag(node, strip_condition=None):
    if isinstance(node, Fragment) or not node.close_tag:
        return im.TextNode('', pos=node.close_tag_pos)

    closetag = im.TextNode(node.close_tag, pos=node.close_tag_pos)
    if strip_condition:
        if strip_condition in {'True', '1'}:
            return im.NullNode(children=[closetag])
        else:
            return im.IfNode(test='not ({})'.format(strip_condition),
                             children=[closetag],
                             pos=node.close_tag_pos)
    return closetag


def _compile_vanilla_element(node):
    container = im.ContainerNode()
    container.append(_compile_open_tag(node))
    container.extend(map(_compile_node, node.children))
    container.append(_compile_close_tag(node))
    return container


def append_directive_nodes(srcnode, directives, container):
    working_node = container
    for directive, data, pos in directives:
        if directive.factory:
            child = directive.factory(**data)
            child.pos = pos
            for _, item, _ in im.walk_tree_pre(child):
                item.pos = pos
            working_node.children.append(child)
            working_node = child
    return working_node


def _compile_element_with_directives(orig, directives, remaining):

    container = working_node = im.ContainerNode()
    inner_directives = [(d, dd, pos) for d, dd, pos in directives if d.inner]
    outer_directives = [(d, dd, pos) for d, dd, pos in directives if not d.inner]

    # Add outer directives
    working_node = append_directive_nodes(orig, outer_directives, container)
    if not isinstance(working_node, im.ContainerNode):
        working_node = im.NullNode()
        container.append(working_node)

    # Add tag
    extra_attrs = []
    strip_condition = None
    for d, dd, pos in directives:
        extras = d.get_extra_attrs(dd, remaining)
        if extras:
            extra_attrs.append(extras)
        strip_condition = (d.get_strip_condition(dd, remaining) or
                           strip_condition)

    working_node.append(_compile_open_tag(remaining,
                                          strip_condition=strip_condition,
                                          extra_attrs=extra_attrs))

    # Add inner directives
    inner_working_node = append_directive_nodes(orig, inner_directives, working_node)

    if not isinstance(inner_working_node, im.ContainerNode):
        inner_working_node = im.NullNode()
        working_node.append(inner_working_node)

    inner_working_node.extend(map(_compile_node, remaining.children))

    working_node.append(_compile_close_tag(remaining,
                                           strip_condition=strip_condition))

    return container



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piglet.position import Position


def test_advance():
    assert Position(1, 1).advance('') == Position(1, 1)
    assert Position(1, 1).advance('f') == Position(1, 2)
    assert Position(1, 3).advance('f\n\na\nccc') == Position(4, 4)
    assert Position(1, 2).advance('f') == Position(1, 3)



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy
from pkg_resources import resource_string

import attr
import parsley

from piglet.exceptions import PigletParseError


@attr.s
class Interpolation(object):
    source = attr.ib()
    value = attr.ib()

    autoescape = True

    def noescape(self):
        ob = copy(self)
        ob.autoescape = False
        return ob


parser_ns = locals()
interp_grammar = resource_string(__name__, 'grammar_interpolate.txt').decode('UTF-8')
interp_parser = parsley.makeGrammar(interp_grammar, parser_ns)


def parse_interpolations(source):
    try:
        return interp_parser(source).text_with_interpolations()
    except parsley.ParseError as e:
        raise PigletParseError(e.formatError())



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ast import (AST,
                 Attribute,
                 For,
                 FunctionDef,
                 Load,
                 Module,
                 Name,
                 Param,
                 Store,
                 With,
                 keyword,
                 parse)
from functools import reduce
from textwrap import dedent
import sys

try:
    from ast import Try as _Try
    _TryExcept = None
except ImportError:
    from ast import TryExcept as _TryExcept
    _Try = None

try:
    from ast import arg
except ImportError:
    arg = None

StoreName = lambda n: Name(id=n, ctx=Store())
LoadName = lambda n: Name(id=n, ctx=Load())


def LoadAttribute(symbol, attr=None):

    if attr is None:
        symbols = symbol.split('.')
        return reduce(LoadAttribute, symbols)

    if isinstance(symbol, AST):
        return Attribute(value=symbol, attr=attr, ctx=Load())
    return Attribute(value=LoadName(symbol), attr=attr, ctx=Load())


def add_arg_default(fn, arg, default):
    fn.args.args.append(arg)
    fn.args.defaults.append(default)


def TryExcept(body, handlers, orelse=None):
    if orelse is None:
        orelse = []
    if _Try:
        return _Try(body=body, handlers=handlers, orelse=orelse, finalbody=[])
    else:
        return _TryExcept(body=body, handlers=handlers, orelse=orelse)


def add_kwarg(call, value):
    """
    Add a **kwarg parameter to a function call node.

    This has different syntax in python3.5+
    """
    if not isinstance(value, AST):
        value = LoadName(value)

    if sys.version_info > (3, 5,):
        call.keywords.append(keyword(arg=None, value=value))
    else:
        call.kwargs = value


def parse_and_strip(pysrc):
    """
    :param pysrc: arbitrary python source code
    :returns: a list of AST nodes, with location information stripped
    """
    pysrc = dedent(pysrc)
    return strip_locations(parse(pysrc)).body


def make_arg(name):
    """
    Shim for differing Python 2 + 3 representations of function args
    """
    if arg is not None:
        return arg(arg=name, annotation=None)
    return Name(id=name, ctx=Param())


def make_kwarg(name):
    """
    Shim for differing Python 2 + 3 representations of function args
    """
    if arg is not None:
        return arg(arg=name, annotation=None)
    return name


def strip_locations(astnode):
    """
    Strip lineno/col_offset annotations from ast.

    This is so that sections of ast generated via parse (which
    has location info embedded) don't mess up the line numbering
    added during compilation
    """
    for n, _ in astwalk(astnode):
        if hasattr(n, 'lineno'):
            del n.lineno
        if hasattr(n, 'col_offset'):
            del n.col_offset
    return astnode


def get_comparison(pysrc):
    """
    :param pysrc: source code of a comparison, eg 'a == b'
    :returns: an AST node suitable for passing to the constructor of ``If``
    """
    return parse_and_strip('if {}: pass'.format(pysrc))[0].test


def astwalk(node, ancestors=None, exclude=lambda ancestors, node, attr: False):
    """
    Walk the AST, generating tuples of ``(node, ancestors)``.

    AST nodes are walked in an order conforming to python's scoping rules:
    nodes are yielded only after all the nodes in their scope have been
    yielded.

    :param exclude: a callable that will be passed each combination of
                    (ancestors, node, attribute_name). If the callable
                    returns True, the given attribute will be skipped in the
                    traversal.

                    For example to exclude for loop contents you would pass::

                        exclude=lambda a, n, attr: \
                            (n.__class__ == For and attr == 'body')
    """
    # List of nodes that indicate a new lexical scope
    CREATES_SCOPE = {Module, FunctionDef, For, With}

    def get_astnodes_in_order(node):
        """
        Generates tuples of ``(astnode, is_child)``

        is_child signals whether the astnode is within the scope of ``node``.
        For example, consider::

            for x in y():
                print(x)

        The print statement is lexically scoped with in the for loop, and
        has access to the symbol ``x``. The call to ``y()`` is still within the
        parent's scope and doesn't have access to ``x``.

        Therefore this function would generate tuples along the lines of this
        pseudo-AST:

            (Call(func='y'), False)
            (Name(id='x', ctx=Store()), False)
            (For(...), False)
            (Call(func='print', args='x'), True)

        """
        # Return fields in this priority order
        field_priority = ['value', 'targets']
        # Fields containing nodes that should be considered children of the
        # current node
        child_attrs = {'body'}

        sort_order = {n: priority for priority, n in enumerate(field_priority)}
        attrs = set(node._fields) - child_attrs
        attrs = sorted(attrs, key=lambda a: sort_order.get(a, len(sort_order)))
        for item in attrs:
            item = getattr(node, item, [])

            if isinstance(item, list):
                for i in item:
                    yield i, False

            elif isinstance(item, AST):
                yield item, False

        yield node, False

        for item in child_attrs:
            item = getattr(node, item, [])

            if isinstance(item, list):
                for i in item:
                    yield i, True

            elif isinstance(item, AST):
                yield item, True

    ancestors = ancestors or tuple()
    scoped = (node.__class__ in CREATES_SCOPE)
    left = []
    right = []
    for n, is_child in get_astnodes_in_order(node):
        if is_child:
            right.append(n)
        else:
            left.append(n)

    for item in left:
        if item is node:
            yield item, ancestors
        else:
            for sub in astwalk(item, ancestors):
                yield sub

    if scoped:
        ancestors = ancestors + (node,)

    for item in right:
        for sub in astwalk(item, ancestors):
            yield sub


def add_locations(node):
    l, c = 1, 1
    for n, ancestors in astwalk(node):
        if hasattr(n, 'lineno'):
            l = n.lineno
        else:
            n.lineno = l
        if hasattr(n, 'column'):
            c = n.col_offset
        else:
            n.col_offset = c
    return node



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class PigletError(Exception):
    """
    Base exception for all piglet templating errors
    """
    lineno = None
    filename = None

    def __str__(self):
        if self.lineno or self.filename:
            return '{} in {}, line {}'.format(self.args[0],
                                            self.filename,
                                            self.lineno)
        return super(PigletError, self).__str__()

    def set_location(self, filename=None, lineno=None):
        if filename is not None and self.filename == self.__class__.filename:
            self.filename = filename
        if lineno is not None and self.lineno == self.__class__.lineno:
            self.lineno = lineno


class PigletParseError(PigletError):
    """
    An error occurred while parsing the input HTML.
    """
    lineno = '(unknown line number)'
    filename = '(unknown filename)'

    def __init__(self, message, filename=None, lineno=None):
        super(PigletParseError, self).__init__(message)
        self.set_location(filename, lineno)


class TemplateNotFound(PigletError):
    """
    Template could not found in specified location
    """


class UndefinedError(PigletError):
    """
    The requested variable is not defined in this context
    """



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piglet import parser
samples = [
    '<html>',
    '<p></p>',
    '&amp;',
    '<a>x</a>',
    '<a><b></b></a>',
    '<img/>',
    '<img />',
    '<a href="foo"></a>',
    '<a  href="foo"></a>',
    '<a  href = "foo" ></a>',
    '<a href="foo" class=\'bar\'>x</a>',
    '<!-- a comment -->',
    '<p>hello <p>world',
    'fish &amp; chips',
    '<p>a</p>',
    '<html:p>a</html:p>',
    '<?php ?>',
    '<!DOCTYPE html>',
    '<!DOCTYPE html><script><![CDATA[ fish & chips ]]></script>',
]

for s in samples:
    print(s)
    print(parser(s).html())



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from os.path import basename, dirname, join as pjoin
from tempfile import NamedTemporaryFile, mkdtemp
import contextlib
import io
import os
import re
import shutil

import pytest

from piglet import TemplateLoader


@pytest.fixture
def tmpfile():
    with NamedTemporaryFile() as tmp:
        yield tmp


@contextlib.contextmanager
def tmpdir():
    t = mkdtemp()
    yield t
    shutil.rmtree(t)


@contextlib.contextmanager
def create_templates(paths):
    """
    Write template files to a temporary directory

    :param paths: list of (path, content) tuples
    :returns: path to temporary directory
    """
    if isinstance(paths, dict):
        paths = paths.items()

    with tmpdir() as d:
        paths = [(pjoin(d, *p.split('/')), content) for p, content in paths]
        for p, c in paths:
            try:
                os.makedirs(dirname(p))
            except OSError:
                pass
            with io.open(p, 'w', encoding='UTF-8') as f:
                f.write(c)

        try:
            yield d
        except Exception:
            import traceback
            traceback.print_exc()
            raise


def test_it_loads_a_template(tmpfile):
    template_text = '<h1>whoa nelly!</h1>'
    tmpfile.write(template_text.encode('UTF-8'))
    tmpfile.flush()
    loader = TemplateLoader([dirname(tmpfile.name)])
    assert loader.load(basename(tmpfile.name)).render({}) == template_text


def test_it_caches_files(tmpfile):
    d, filename = os.path.split(tmpfile.name)
    loader = TemplateLoader([d], auto_reload=True)
    with open(tmpfile.name, 'wb') as f:
        f.write(b't1')
    t = loader.load(filename)
    assert loader.load(filename) is t
    with open(tmpfile.name, 'wb') as f:
        f.write(b't2')
    assert loader.load(filename) is not t


def test_it_searches_path():

    template_text = '<h1>whoa nelly!</h1>'

    with tmpdir() as t1, tmpdir() as t2, tmpdir() as t3:
        loader = TemplateLoader([t1, t2, t3])
        for d in [t1, t2, t3]:
            with NamedTemporaryFile(dir=d) as f:
                f.write(template_text.encode('utf-8'))
                f.flush()
                t = loader.load(basename(f.name))
                assert t.render({}) == template_text
                assert t.filename == f.name


def test_it_loads_from_inside_template_same_dir():
    with create_templates([
        ('a/t1', u't1 <py:block name="b">hello</py:block>'),
        ('a/t2', u'<a py:extends="t1"><py:block name="b">t2</py:block></a>')]
    ) as d:
        loader = TemplateLoader([d])
        assert loader.load('a/t2').render({}) == 't1 t2'


def test_it_caches_python_source():
    with create_templates([('t1', u'whoa nelly!')]) as d:
        loader = TemplateLoader([d], cache_dir=d)
        loader.load('t1')
        cache_file = os.path.join(
            d,
            next(f for f in os.listdir(d) if f.endswith('.py'))
        )
        with io.open(cache_file, 'r') as f:
            cached = f.read()
        assert re.compile(r"yield u?'whoa nelly!'").search(cached) is not None

        loader._cache.clear()
        loader._level2_cache.clear()

        with io.open(cache_file, 'w') as f:
            f.write(cached.replace('nelly', 'smelly'))
        assert loader.load('t1').render({}) == 'whoa smelly!'


def test_it_loads_from_across_template_dirs():
    templates = create_templates([
        ('a/t1', u't1 <py:block name="b">hello</py:block>'),
        ('b/t2', u'<a py:extends="t1"><py:block name="b">t2</py:block></a>')])

    with templates as d:
        loader = TemplateLoader([pjoin(d, 'a'), pjoin(d, 'b')])
        assert loader.load('t2').render({}) == 't1 t2'

        TemplateLoader([pjoin(d, 'b'), pjoin(d, 'a')])
        assert loader.load('t2').render({}) == 't1 t2'


def test_it_loads_relative():
    templates = create_templates([
        ('a/t1', 'a-t1 <py:block name="b">hello</py:block>'),
        ('b/t1', 'b-t1 <py:block name="b">hello</py:block>'),
        ('b/t2', '<a py:extends="./t1"><py:block name="b">t2</py:block></a>')])
    with templates as d:
        loader = TemplateLoader([pjoin(d, 'a'), pjoin(d, 'b')])
        assert loader.load('t2').render({}) == 'b-t1 t2'


def test_it_loads_absolute():
    templates = create_templates([
        ('a/t1', 'a-t1 <py:block name="b">hello</py:block>'),
        ('b/t1', 'b-t1 <py:block name="b">hello</py:block>'),
        ('b/t2', '<a py:extends="t1"><py:block name="b">t2</py:block></a>')])
    with templates as d:
        loader = TemplateLoader([pjoin(d, 'a'), pjoin(d, 'b')])
        assert loader.load('t2').render({}) == 'a-t1 t2'


def test_it_doesnt_load_current_template():
    templates = create_templates([
        ('a/page', 'a-page <py:extends href="layout"></py:extends>'),
        ('a/layout', 'a-layout <py:extends href="layout"></py:extends>'),
        ('b/layout', 'b-layout')])
    with templates as d:
        loader = TemplateLoader([pjoin(d, 'a'), pjoin(d, 'b')])
        assert loader.load('page').render({}) == 'a-page a-layout b-layout'



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gettext import NullTranslations
import marshal

from piglet import compilexml
from piglet import compilett
from piglet.parse import parse_html
from piglet import compile as pigletcompile
from piglet.exceptions import PigletParseError
from piglet.runtime import data as rtdata, munge_exception_messages

ustr = type(u'')


class BaseTemplate:

    filename = None
    loader = None
    template_module = None
    codeob = None
    intermediate = None

    def __init__(self,
                 source=None,
                 filename='<string>',
                 loader=None,
                 translations_factory=None,
                 save_pysrc_to=None,
                 codeob=None):
        self.filename = filename
        self.loader = loader
        if translations_factory:
            self.translations_factory = translations_factory
        try:
            if codeob is None:
                self.intermediate = self.compile_intermediate(source)
                codeob = pigletcompile.compile_to_codeob(self.intermediate,
                                                         filename,
                                                         saveas=save_pysrc_to)
            self.codeob = codeob
            self.template_module = self._get_module()
        except PigletParseError as e:
            e.filename = filename
            raise
        assert self.template_module is not None
        self.root_fn = self.template_module.__piglet_root__

    @classmethod
    def from_pysrc(cls,
                    pysrc,
                    pysrc_filename='<string>',
                    filename='<string>',
                    loader=None,
                    translations_factory=None):

        return cls(filename=filename,
                   codeob=compile(pysrc, pysrc_filename, 'exec'),
                   loader=loader,
                   translations_factory=translations_factory)

    def __repr__(self):
        return "<{} {!r}>".format(type(self).__name__, self.filename)

    def __call__(self, context, *args, **kwargs):
        if not hasattr(rtdata, 'context'):
            rtdata.context = []
        rtdata.context.append(context)
        rtdata.exception_locations = []
        translations = self.translations_factory()
        context.update({'_': translations.gettext,
                        'gettext': translations.gettext,
                        'ngettext': translations.ngettext})
        content = munge_exception_messages(self.root_fn(*args, **kwargs),
                                           context)
        try:
            for s in content:
                yield s
        finally:
            rtdata.context.pop()

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['template_module']
        del d['root_fn']
        d['codeob'] = marshal.dumps(self.codeob)
        return d

    def __setstate__(self, state):
        state['codeob'] = marshal.loads(state['codeob'])
        self.__dict__.update(state)
        self.__dict__['template_module'] = self._get_module()
        self.root_fn = self.template_module.__piglet_root__

    def _get_module(self):
        return pigletcompile.compile_to_module(
            self.codeob,
            self.filename,
            bootstrap={'__piglet_template': self})

    def compile_intermediate(self):
        raise NotImplementedError()

    def render(self, context, *args, **kwargs):
        return ''.join(map(ustr, self(context, *args, **kwargs)))

    def get_pysrc(self):
        """
        Return the python source for the template file
        This may return None in the case that the template was not
        compiled from source.
        """
        if self.intermediate is None:
            return None
        return pigletcompile.compile_to_source(self.intermediate)

    def __getattr__(self, name):
        """
        Expose template functions as attributes, eg:

            >>> t = Template('<py:def function="greeting">Hello!</py:def>')
            >>> print(t.greeting())
            'Hello!'
        """
        return getattr(self.template_module, name)

    def translations_factory(self):
        return NullTranslations()


class HTMLTemplate(BaseTemplate):
    def compile_intermediate(self, src):
        parsed = parse_html(src)
        return compilexml.compile_intermediate(parsed)


class TextTemplate(BaseTemplate):
    def compile_intermediate(self, src):
        parsed = compilett.parse_tt(src)
        return compilett.compile_intermediate(parsed)


Template = HTMLTemplate



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os.path import join as pjoin
import io
import pkg_resources

from piglet import compilexml as cx
from piglet import intermediate as im
from piglet.compile import compile_to_source
from piglet.parse import parse_html
from piglet.position import Position as P
from piglet.template import Template
from piglet.test_loader import create_templates

from mock import Mock

EXTRACT_KEYWORDS = {'_', 'gettext', 'ngettext'}


def toim(s):
    return cx.compile_intermediate(parse_html(s)).children


def tosrc(s):
    return compile_to_source(cx.compile_intermediate(parse_html(s)))


class TestI18N:
    def test_it_compiles_i18n_message(self):
        n = toim('<div i18n:message="" i18n:comment="translate me">foo</div>')
        assert n == [
            im.TextNode(pos=P(1, 1), content='<div>'),
            im.TranslationNode(pos=P(1, 20),
                               message='',
                               comment='translate me',
                               children=[im.TextNode(pos=P(1, 50),
                                                     content='foo')]),
            im.TextNode(pos=P(1, 53), content='</div>'),
        ]
        assert n[1].get_msgstr() == 'foo'

    def test_it_compiles_i18n_message_with_custom_message(self):
        n = toim('<div i18n:message="bar">foo</div>')
        assert n == [
            im.TextNode(pos=P(1, 1), content='<div>'),
            im.TranslationNode(pos=P(1, 20),
                               message='bar',
                               comment=None,
                               children=[im.TextNode(pos=P(1, 25),
                                                     content='foo')]),
            im.TextNode(pos=P(1, 28), content='</div>'),
        ]
        assert n[1].get_msgstr() == 'bar'

    def test_it_substitutes_dynamic_placeholders(self):
        n = toim('<div i18n:message="">foo $x <py:if test="x">1</py:if></div>')
        assert n[1].get_msgstr() == 'foo ${x} ${dynamic.1}'

    def test_it_substitutes_named_placeholders(self):
        n = toim('<div i18n:message="">foo <b i18n:name="x">$x</b></div>')
        assert n[1].get_msgstr() == 'foo ${x}'

        n = toim('<div i18n:message="">foo <i18n:s name="x" expr="x"/></div>')
        assert n[1].get_msgstr() == 'foo ${x}'
        interpolation = n[1].children[1].children[0]
        assert isinstance(interpolation, im.InterpolateNode)
        assert interpolation.value == 'x'

    def test_it_processes_placeholder_values(self):
        t = Template('<div i18n:message="">foo <b i18n:name="x">$x</b></div>')
        assert t.render({'x': 'baa'}) == '<div>foo <b>baa</b></div>'

        t.translations_factory = lambda: Mock(
            gettext=Mock(return_value='${x} sheep'))

        assert t.render({'x': 'baa'}) == '<div><b>baa</b> sheep</div>'

        t.translations_factory = lambda: Mock(
            gettext=Mock(return_value='BINGO!'))
        assert t.render({'x': 'baa'}) == '<div>BINGO!</div>'

    def test_it_normalizes_whitespace(self):
        t = Template('<div i18n:message="">\nfoo \n \n bar </div>')
        assert t.render({}) == '<div>foo bar</div>'


class TestExtractor:

    def get_extractor(self, name='piglet'):
        for ep in pkg_resources.iter_entry_points('babel.extractors'):
            if ep.dist.project_name == 'piglet' and ep.name == name:
                return ep.load()
        assert False, "No setuptools entry point exposed"

    def test_it_extracts_from_i18n_directives(self):
        with create_templates([
            ('a', u'<a i18n:message="foo">bar</a>'),
            ('b', u'<a i18n:message="">bar</a>'),
            ('c', u'<a i18n:message="">x is ${x}</a>'),
            ('d', u'<a i18n:message="">y is <a i18n:name="why">${x}</a></a>'),
            ('e', u'<a i18n:message="">${ x }</a>'),
        ]) as d:
            file_messages = [('a', [(1, '_', 'foo', [])]),
                             ('b', [(1, '_', 'bar', [])]),
                             ('c', [(1, '_', 'x is ${x}', [])]),
                             ('d', [(1, '_', 'y is ${why}', [])]),
                             ('e', [(1, '_', '${x}', [])]),
                             ]
            e = self.get_extractor()
            for f, expected in file_messages:
                with io.open(pjoin(d, f), 'r', encoding='utf-8') as f:
                    assert list(e(f, EXTRACT_KEYWORDS, [], {})) == expected

    def test_it_extracts_from_i18n_attributes_on_py_directives(self):
        with create_templates([
            ('a', (u'<py:if test="1" i18n:message="">one</py:if>'
                   u'<py:else i18n:message="">two</py:else>')),
        ]) as d:
            file_messages = [('a', [(1, '_', 'one', []),
                                    (1, '_', 'two', [])]),
                             ]
            e = self.get_extractor()
            for f, expected in file_messages:
                with io.open(pjoin(d, f), 'r', encoding='utf-8') as f:
                    assert list(e(f, EXTRACT_KEYWORDS, [], {})) == expected

    def test_it_extracts_from_interpolations(self):

        with create_templates([
            ('a', u'<a title="${_(\'foo\')}"></a>'),
            ('b', u'${gettext("bar")}'),
            ('c', u'${ngettext("fish", "fishes", n)}'),
        ]) as d:
            file_messages = [('a', [(1, '_', 'foo', [])]),
                             ('b', [(1, 'gettext', 'bar', [])]),
                             ('c', [(1, 'ngettext', ('fish', 'fishes'), [])]),
                             ]
            e = self.get_extractor()
            for f, expected in file_messages:
                with io.open(pjoin(d, f), 'r', encoding='utf-8') as f:
                    assert list(e(f, EXTRACT_KEYWORDS, [], {})) == expected



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import chain
from pkg_resources import resource_string
import re

import attr
import parsley


#: Grammar to parse semicolon separated values in WithNodes
ssv_grammar = resource_string(__name__, 'grammar_ssv.txt').decode('UTF-8')
ssv_parser = parsley.makeGrammar(ssv_grammar, {})


@attr.s
class IntermediateNode(object):
    """
    An intermediate representation of the template structure.

    A template such as::

        <py:for each="i, x in enumerate(xs)">
            <a py:if="x.href" href="$x.href">
                link to ${x.name}
            </a>
        </py:for>

    Could be modelled as::

        ForNode(
            each="i, x in enumerate(xs)",
            children=[
                IfNode(expr='x.href',
                       children=[TextNode('<a href="'),
                                 InterpolateNode('x.href'),
                                 TextNode('">link to "),
                                 InterpolateNode('x.name'),
                                 TextNode('</a>')])])
    """

    pos = attr.ib(default=None)
    children = []

    @classmethod
    def factory(cls, *args, **kwargs):
        raise NotImplementedError(cls)

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        return self

    def find(self, cls=object, test=lambda node: True):
        """
        Walk this node and all children, yielding matching nodes

        :param cls: Match only nodes of this class
        :param test: Match only nodes for which this predicate returns True
        """
        for parent, node, idx in walk_tree_pre(self):
            if isinstance(node, cls) and test(node):
                yield node

    @property
    def all_children(self):
        """
        Some nodes may have multiple lists of children (eg an IfNode
        has both ``children`` and ``else_``. The ``all_children``
        property should return an iterable over all child nodes for the
        purpose of tree walking
        """
        return self.children


@attr.s
class ContainerNode(IntermediateNode):
    """
    A generic container node
    """
    children = attr.ib(attr.Factory(list))

    @classmethod
    def factory(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def append(self, child):
        """
        Add child to this node's list of children.
        """
        self.children.append(child)

    def extend(self, children):
        self.children.extend(children)

    def tip(self):
        """
        Find the newest tip of the tree
        """
        if not self.children:
            return self

        if isinstance(self.children[-1], ContainerNode):
            return self.children[-1].tip()

        return self.children[-1]


class DirectiveNode(object):

    @classmethod
    def factory(cls, **kwargs):
        return cls(**kwargs)


class RootNode(ContainerNode):
    "Container node for an entire document"

    @classmethod
    def factory(cls):
        return cls()


@attr.s(init=False)
class TextNode(IntermediateNode):

    content = attr.ib(default=None)

    def __init__(self, content=None, pos=None):
        self.content = content
        self.pos = pos

    @classmethod
    def factory(cls, content=None, pos=None):
        return cls(content, pos)


@attr.s
class BlockNode(ContainerNode):
    """
    Models a py:block directive
    """
    name = attr.ib(default=None)

    @classmethod
    def factory(cls, name):
        return cls(name=name)


@attr.s
class IfNode(DirectiveNode, ContainerNode):
    """
    Models a py:if ... py:else directive
    """
    test = attr.ib(default=None)
    else_ = attr.ib(default=None)

    @property
    def all_children(self):
        if self.else_:
            return chain(self.children, [self.else_])
        return self.children


class ElseNode(DirectiveNode, ContainerNode):
    """
    The else part of a ``py:if ... py:else`` directive.
    """


@attr.s
class DefNode(DirectiveNode, ContainerNode):
    """
    Models a py:def directive
    """
    function = attr.ib(default=None)


@attr.s
class ExtendsNode(ContainerNode):
    """
    Models a py:extends directive
    """
    href = attr.ib(default=None)

    @classmethod
    def factory(cls, href):
        return cls(href=href)


@attr.s
class IncludeNode(DirectiveNode, IntermediateNode):
    """
    Models a py:include directive
    """
    href = attr.ib(default=None)


@attr.s
class ChooseNode(DirectiveNode, ContainerNode):
    """
    Models a py:choose directive.

    A choose node may have When, Otherwise and Text children.
    If :attribute:`ChooseNode.test` contains an expression, the value of that
    expression is is compared to any values contained in
    :attribute:`WhenNode.test` children. The first matching :class:`When` node
    is rendered and the others are dropped.

    If :attribute:`ChooseNode.test` is empty, each contained
    :attribute:`WhenNode.test` is evaluated as a boolean expression and the
    first truthful result is rendered.

    If no :class:`WhenNode` is rendered, any :class:`OtherwiseNode` directives
    will be rendered.
    """
    test = attr.ib(default=None)


@attr.s
class WhenNode(DirectiveNode, ContainerNode):
    """
    Models a py:when directive, and must be contained with a
    :class:`ChooseNode` directive.
    """
    test = attr.ib(default=None)


class OtherwiseNode(DirectiveNode, ContainerNode):
    pass


@attr.s
class ForNode(DirectiveNode, ContainerNode):
    """
    Models a py:for directive.

    The expression in :attribute:`ForNode.each` must be in the form
    `<target> in <iterator>`, and is used to generate a python for loop.
    """
    each = attr.ib(default=None)


@attr.s
class WithNode(DirectiveNode, ContainerNode):
    """
    Models a py:with directive.

    The expression in :attribute:`WithNode.vars` must be semicolon separated
    list of variable assignments, for example:

        WithNode(vars='x = 1; y = x * 2')

    The assigned variables will be available only within the scope of the
    directive
    """
    vars = attr.ib(default=[])

    def get_pairs(self):
        """
        Return the configured variables as a list of (target, expr) pairs.
        """
        if isinstance(self.vars, list):
            return self.vars

        values = ssv_parser(self.vars).values()
        values = [tuple(s.strip() for s in str(item).split('=', 1))
                    for item in values]
        return values


@attr.s
class InterpolateNode(DirectiveNode, IntermediateNode):
    """
    Renders a python expression interpolation
    """
    value = attr.ib(default=None)
    autoescape = attr.ib(default=True)


class NullNode(ContainerNode):
    """
    Used to account for text that should not appear in the compiled template,
    but still needs a node to keep the line numbering correct.
    """


@attr.s
class ImportNode(DirectiveNode, ContainerNode):
    href = attr.ib(default=None)
    alias = attr.ib(default=None)


@attr.s
class InlineCodeNode(DirectiveNode, ContainerNode):
    pysrc = attr.ib(default=None)


@attr.s
class TranslationNode(DirectiveNode, ContainerNode):
    message = attr.ib(default=None)
    comment = attr.ib(default=None)

    def get_msgstr(self):

        if self.message:
            return self.message

        s = []
        for name, item in self.named_children():
            if name is None:
                s.append(item.content)
            else:
                s.append('${{{}}}'.format(name))
        s = ''.join(s)
        return re.sub('[ \t\r\n]+', ' ', s, re.S).strip()

    def named_children(self):
        """
        Return a tuples of ('placeholder_name', imnode).
        For TextNode children, placeholder_name will be None.
        """
        dyn_index = 1
        for item in self.children:
            if isinstance(item, TextNode):
                yield (None, item)
            elif isinstance(item, TranslationPlaceholder):
                yield (item.name, item)
            elif isinstance(item, InterpolateNode):
                yield (item.value.strip(), item)
            else:
                yield ('dynamic.{}'.format(dyn_index), item)
                dyn_index += 1


@attr.s
class TranslationPlaceholder(DirectiveNode, ContainerNode):
    name = attr.ib(default=None)


@attr.s
class Call(DirectiveNode, ContainerNode):
    """
    A python function call
    """
    function = attr.ib(default=None)


@attr.s
class CallKeyword(DirectiveNode, ContainerNode):
    """
    A keyword argument to a function :class:`Call`
    that may contain an arbitrary template snippet
    """
    name = attr.ib(default=None)


def _walk_tree(order, n, parent, pos):
    assert order in {'pre', 'post'}
    if order == 'pre':
        yield parent, n, pos
        children = enumerate(n.all_children)
    else:
        children = reversed(list(enumerate(n.all_children)))

    for index, sub in children:
        for item in _walk_tree(order, sub, n, index):
            yield item
    if order == 'post':
        yield parent, n, pos


def walk_tree_post(n, parent=None, pos=0):
    """
    Walk the intermediate tree in a post order traversal.

    Yields tuples of (parent, node, index).

    A post order traversal is chosen so that nodes may be deleted
    or merged without affecting the subsequent traversal.
    """
    return _walk_tree('post', n, parent, pos)


def walk_tree_pre(n, parent=None, pos=0):
    """
    Walk the intermediate tree in a pre order traversal.

    Yields tuples of (parent, node, index).
    """
    return _walk_tree('pre', n, parent, pos)

walk_tree = walk_tree_pre



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from functools import wraps, partial
import sys
import threading

from markupsafe import Markup, escape_silent
from piglet.exceptions import PigletError, UndefinedError

try:
    import builtins
except ImportError:
    import __builtin__ as builtins

Markup = Markup
escape = escape_silent
partial = partial
builtins = builtins
exc_info = sys.exc_info

data = threading.local()


def load(relative_to, src):
    loader = relative_to.loader
    if loader is None:
        raise Exception("Template has no loader attached. Ensure template "
                        "was loaded via piglet.TemplateLoader.load.")
    template = loader.load(src, relative_to=relative_to)
    if template is relative_to:
        raise PigletError("Template {!r} may not include itself"
                          .format(template))
    return template


class Undefined(object):

    __slots__ = ('name',)

    def __init__(self, name):
        object.__setattr__(self, 'name', name)

    def raise_undefined_error(self, *args, **kwargs):
        raise UndefinedError(self.name)

    __call__ = raise_undefined_error
    __hash__ = raise_undefined_error
    __sizeof__ = raise_undefined_error
    __repr__ = raise_undefined_error
    __str__ = raise_undefined_error
    __dir__ = raise_undefined_error
    __format__ = raise_undefined_error
    __subclasses__ = raise_undefined_error
    __floor__ = raise_undefined_error
    __trunc__ = raise_undefined_error
    __ceil__ = raise_undefined_error
    __cmp__ = raise_undefined_error
    __lt__ = raise_undefined_error
    __gt__ = raise_undefined_error
    __le__ = raise_undefined_error
    __ge__ = raise_undefined_error
    __eq__ = raise_undefined_error
    __ne__ = raise_undefined_error
    __getitem__ = raise_undefined_error
    __setitem__ = raise_undefined_error
    __delitem__ = raise_undefined_error
    __contains__ = raise_undefined_error
    __len__ = raise_undefined_error
    __iter__ = raise_undefined_error
    __getslice__ = raise_undefined_error
    __setslice__ = raise_undefined_error
    __reversed__ = raise_undefined_error
    __missing__ = raise_undefined_error
    __enter__ = raise_undefined_error
    __exit__ = raise_undefined_error
    __neg__ = raise_undefined_error
    __pos__ = raise_undefined_error
    __invert__ = raise_undefined_error
    __add__ = raise_undefined_error
    __sub__ = raise_undefined_error
    __mul__ = raise_undefined_error
    __div__ = raise_undefined_error
    __floordiv__ = raise_undefined_error
    __mod__ = raise_undefined_error
    __divmod__ = raise_undefined_error
    __lshift__ = raise_undefined_error
    __rshift__ = raise_undefined_error
    __and__ = raise_undefined_error
    __xor__ = raise_undefined_error
    __or__ = raise_undefined_error
    __pow__ = raise_undefined_error
    __complex__ = raise_undefined_error
    __int__ = raise_undefined_error
    __float__ = raise_undefined_error
    __index__ = raise_undefined_error
    __coerce__ = raise_undefined_error
    __get__ = raise_undefined_error
    __set__ = raise_undefined_error
    __delete__ = raise_undefined_error
    __reduce__ = raise_undefined_error
    __reduce_ex__ = raise_undefined_error
    __getinitargs__ = raise_undefined_error
    __getnewargs__ = raise_undefined_error
    __getstate__ = raise_undefined_error
    __setstate__ = raise_undefined_error
    __unicode__ = raise_undefined_error
    __long__ = raise_undefined_error
    __oct__ = raise_undefined_error
    __hex__ = raise_undefined_error
    __nonzero__ = raise_undefined_error
    __truediv_ = raise_undefined_error
    __rtruediv__ = raise_undefined_error
    __bool__ = raise_undefined_error
    __next__ = raise_undefined_error
    __getattr__ = raise_undefined_error
    __setattr__ = raise_undefined_error
    __prepare__ = raise_undefined_error
    __instancecheck__ = raise_undefined_error
    __subclasscheck__ = raise_undefined_error


class FlatOutput:

    def __init__(self, iterable):
        self.iterable = iterable

    def __iter__(self, *args, **kwargs):
        return self.iterable

    def __str__(self):
        return ''.join(self.iterable)

    __html__ = __unicode__ = __str__


def flatten(fn):
    @wraps(fn)
    def flatten(*args, **kwargs):
        return FlatOutput(fn(*args, **kwargs))
    return flatten


def munge_exception_messages(iterable, context):
    context['__piglet_exc_locations'] = []
    try:
        for item in iterable:
            yield item
    except Exception:
        cls, exc, tb = sys.exc_info()
        if getattr(exc, '__piglet_is_annotated__', False):
            raise

        saved_msg = str(exc)

        def annotated_message(self):
            locations = context['__piglet_exc_locations']
            tb = ('{}"{}", line {}'.format('    ' * ix, f, l)
                    for ix, (f, l) in enumerate(locations))

            return ('{}\n'
                    'Exception originated in code starting at {}'
                    .format(saved_msg, '\n'.join(tb)))

        AnnotatedException = type(
            cls.__name__,
            (cls,),
            {'__str__': annotated_message,
             '__module__': cls.__module__,
             '__traceback__': tb,
             '__piglet_is_annotated__': True}
        )

        newexc = AnnotatedException.__new__(AnnotatedException)
        newexc.__init__(*exc.args)
        newexc.__dict__ = newexc.__dict__
    else:
        return
    raise newexc


@flatten
def nocontent(*args, **kwargs):
    """
    Used as default value for ``super`` argument in py:block functions.
    """
    return iter([])


def get_super(bases, current_template, name, default=nocontent):
    """
    Return the super block function from the stack of bases::

        >>> bases = [template_a, template_b]
        >>> get_super(bases, 'page')

    This would return a partial function application looking like this::

        partial(template_b.page, bases=[template_a])
    """
    def super(*args, **kwargs):
        try:
            eligible_bases = bases[:bases.index(current_template)]
        except ValueError:
            eligible_bases = bases
        while eligible_bases:
            try:
                super_fn = getattr(eligible_bases.pop(), name)
            except AttributeError:
                continue
            return super_fn(__piglet_bases=eligible_bases, *args, **kwargs)
        return default()
    return super



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import attr


@attr.s(slots=True, frozen=True)
class Position(object):
    line = attr.ib()
    char = attr.ib()

    def advance(self, text):
        return self.__class__(*update_pos(self, text))


def update_pos(oldpos, s):
    """
    Update a line, char position as if the cursor has moved through text ``s``.
    """
    l, c = attr.astuple(oldpos)
    l += s.count('\n')
    last_newline = s.rfind('\n')
    if last_newline < 0:
        c = c + len(s)
    else:
        c = len(s) - last_newline
    return (l, c)



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from hypothesis import given, strategies as s

from piglet import compilett, TemplateLoader, TextTemplate
from piglet.test_loader import create_templates


def toim(s):
    return compilett.compile_intermediate(compilett.parse_tt(s)).children


def test_it_compiles_if():
    tt = TextTemplate('{% if x > 1 %}yes{%end%}')
    assert tt.render({'x': 2}) == 'yes'
    assert tt.render({'x': 1}) == ''


def test_it_compiles_if_else():
    tt = TextTemplate('{% if x > 1 %}yes{% else %}no{%end%}')
    assert tt.render({'x': 2}) == 'yes'
    assert tt.render({'x': 1}) == 'no'


def test_it_compiles_with():
    tt = TextTemplate('{% with x = "1"; y = "2"; z=y %}${x + y + z}{% end %}')
    assert tt.render({}) == '122'


def test_it_compiles_trans():
    n = toim('{% trans %}foo {% transname x %}$x{%end%}{% end %}')
    assert n[0].get_msgstr() == 'foo ${x}'


def test_it_doesnt_html_escape():
    tt = TextTemplate('$x')
    assert tt.render({'x': '&'}) == '&'


def test_it_strips_whitespace_when_directive_tag_alone_on_line():
    tt = TextTemplate('A \n\n {% if 1 %} \n B \n  {%end%} \n C\n')
    assert tt.render({}) == 'A \n\n B \n C\n'


@given(s.text(alphabet='\n \t'),
       s.text(alphabet='\n \t'),
       s.text(alphabet='\n \t'),
       s.text(alphabet='\n \t'),
       s.sampled_from({'', '-', '+'}),
       s.sampled_from({'', '-', '+'}),
       s.sampled_from({'', '-', '+'}),
       s.sampled_from({'', '-', '+'}),
       )
def test_it_applies_whitespace_control(s1, s2, s3, s4, w1, w2, w3, w4):

    source = ('A{s1}{{%{w1} if 1 {w2}%}}{s2}B{s3}{{%{w3}end{w4}%}}{s4}C'
              .format(**locals()))
    tt = TextTemplate(source)

    expected = 'A'
    if w1 == '+':
        expected += s1
    elif w1 == '':
        expected += re.sub(r'\n[ \t]*\Z', r'\n', s1)
    if w2 == '+':
        expected += s2
    elif w2 == '':
        expected += re.sub(r'\A[ \t]*\n', r'', s2)

    expected += 'B'

    if w3 == '+':
        expected += s3
    elif w3 == '':
        expected += re.sub(r'\n[ \t]*\Z', r'\n', s3)
    if w4 == '+':
        expected += s4
    elif w4 == '':
        expected += re.sub(r'\A[ \t]*\n', r'', s4)

    expected += 'C'
    result = tt.render({})
    assert result == expected


def test_it_compiles_extends_and_block():
        with create_templates([
            ('a', u'A {% block foo %}foo{% end %}'),
            ('b', u'{% extends "a" %}{% block foo %}bar{% end %}{% end %}')
        ]) as d:
            loader = TemplateLoader([d], template_cls=TextTemplate)
            assert loader.load('a').render({}) == 'A foo'
            assert loader.load('b').render({}) == 'A bar'



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pkg_resources import resource_string
from collections import OrderedDict

import attr
import parsley

from piglet.exceptions import PigletParseError
from piglet.position import Position


@attr.s
class ParseItem(object):
    pos = attr.ib(default=None)

    def set_pos(self, pos):
        self.pos = pos
        return pos.advance(self.source)

    @property
    def source(self):
        raise NotImplementedError()


@attr.s
class OpenTag(ParseItem):

    qname = attr.ib(default=None)

    #: Whitespace between the name and the first attribute
    space = attr.ib(default=None)
    attrs = attr.ib(default=attr.Factory(OrderedDict),
                    convert=OrderedDict)

    end = attr.ib(default='>')

    def set_pos(self, pos):
        cursor = pos.advance('<{}{}'.format(self.qname, self.space))
        for item in self.attrs.values():
            cursor = item.set_pos(cursor)
        return super(OpenTag, self).set_pos(pos)

    @property
    def source(self):
        attrs = ''.join(v.source for v in self.attrs.values())
        return '<{}{}{}{}'.format(self.qname, self.space, attrs, self.end)


@attr.s
class Attribute(ParseItem):
    name = attr.ib(default=None)
    value = attr.ib(default=None)
    quote = attr.ib(default=None)
    space1 = attr.ib(default=None)
    space2 = attr.ib(default=None)
    space3 = attr.ib(default=None)

    #: The start position of the attribute's value
    value_pos = attr.ib(default=None)

    def __init__(self, *args, **kwargs):
        super(Attribute, self).__init__(*args, **kwargs)

    def set_pos(self, pos):
        self.value_pos = pos.advance('{0.name}{0.space1}={0.space2}{0.quote}'
                                     .format(self))
        return super(Attribute, self).set_pos(pos)

    @property
    def source(self):
        return ('{0.name}{0.space1}='
                '{0.space2}{0.quote}{0.value}{0.quote}{0.space3}'.format(self))


@attr.s
class CloseTag(ParseItem):
    qname = attr.ib(default=None)

    @property
    def source(self):
        return '</{}>'.format(self.qname)


class OpenCloseTag(OpenTag):

    @property
    def source(self):
        attrs = ''.join(v.source for v in self.attrs.values())
        return '<{}{}{}>'.format(self.qname,
                                 self.space,
                                 attrs)


@attr.s
class Comment(ParseItem):
    content = attr.ib(default=None)

    @property
    def source(self):
        return '<!--{}-->'.format(self.content)


@attr.s
class Text(ParseItem):
    content = attr.ib(default=None)
    cdata = attr.ib(default=False)

    @property
    def source(self):
        return self.content


@attr.s
class Entity(ParseItem):
    reference = attr.ib(default=None)

    @property
    def source(self):
        return self.reference


@attr.s
class PI(ParseItem):
    target = attr.ib(default=None)
    content = attr.ib(default=None)

    @property
    def source(self):
        return '<?{}{}?>'.format(self.target, self.content)


@attr.s
class Declaration(ParseItem):
    content = attr.ib(default=None)

    @property
    def source(self):
        return '<!{}>'.format(self.content)


@attr.s
class CDATA(ParseItem):
    content = attr.ib(default=None)

    @property
    def source(self):
        return '<![CDATA[{}]]>'.format(self.content)


parser_ns = locals()
html_grammar = resource_string(__name__, 'grammar.txt').decode('UTF-8')
html_parser = parsley.makeGrammar(html_grammar, parser_ns)


def parse_html(source):
    try:
        return add_positions(list(flatten(html_parser(source).html())))
    except parsley.ParseError as e:
        raise PigletParseError(e.formatError())


def flatten(parse_result):
    for item in parse_result:
        if isinstance(item, list):
            for i2 in flatten(item):
                yield i2
        else:
            yield item


def add_positions(parse_result):
    """
    Add position annotations to each parsed item
    """
    cursor = Position(1, 1)
    for item in parse_result:
        cursor = item.set_pos(cursor)
    return parse_result



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = '0.4'

from piglet.loader import TemplateLoader
from piglet.template import HTMLTemplate, TextTemplate
from piglet.exceptions import PigletError, PigletParseError, TemplateNotFound

from parsley import ParseError

# Monkey patch to avoid 'TypeError: unhashable type' internal errors when
# running under pytest.
ParseError.__hash__ = Exception.__hash__



#!/usr/bin/env python
# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from setuptools import setup

VERSIONFILE = "piglet/__init__.py"


def get_version():
    with open(VERSIONFILE, 'rb') as f:
        return re.search("^__version__\s*=\s*['\"]([^'\"]*)['\"]",
                           f.read().decode('UTF-8'), re.M).group(1)


def read(*path):
    """
    Return content from ``path`` as a string
    """
    with open(os.path.join(os.path.dirname(__file__), *path), 'rb') as f:
        return f.read().decode('UTF-8')


setup(name='piglet',
      version=get_version(),
      description='Fast, HTML aware templating engine',
      long_description=read('README.rst') + '\n\n' + read('CHANGELOG.rst'),
      url='https://bitbucket.org/ollyc/piglet',
      author='Oliver Cope',
      author_email='oliver@redgecko.org',
      license='Apache',
      keywords=['jinja2', 'genshi', 'kajiki', 'mako', 'kid', 'template',
                'templating'],
      install_requires=['Parsley',
                        'attrs',
                        'astunparse',
                        'markupsafe',
                        ],
      packages=['piglet'],
      include_package_data=True,
      entry_points={'babel.extractors': ['piglet=piglet.i18n:extract']}
      )



# Copyright 2016 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from pkg_resources import resource_string
import re

import attr
import parsley

from piglet import parse
from piglet import compilexml as cx
from piglet import intermediate as im
from piglet.exceptions import PigletParseError


@attr.s
class Statement(parse.ParseItem):
    name = attr.ib(default=None)
    args = attr.ib(default=None)
    space1 = attr.ib(default=None)
    space2 = attr.ib(default=None)
    end_tag = attr.ib(default=None)
    end_tag_pos = attr.ib(default=None)
    children = attr.ib(default=attr.Factory(list))
    ws_before = attr.ib(default=None)
    ws_after = attr.ib(default=None)

    @property
    def source(self):
        return '{{%{0.space1}{0.name}{0.space2}{0.args}%}}'.format(self)

    @property
    def raw_args(self):
        return ''.join(a.source if isinstance(a, QuotedString) else a
                       for a in self.args)

    def is_open_tag(self):
        return not self.name.startswith('end')

    def is_close_tag(self):
        return self.name.startswith('end')


@attr.s
class QuotedString(parse.ParseItem):
    quote = attr.ib(default=None)
    content = attr.ib(default=None)

    @property
    def source(self):
        return '{0.quote}{0.content}{0.quote}'.format(self)


@attr.s
class Container(object):
    children = attr.ib(default=attr.Factory(list))


parser_ns = {'Statement': Statement,
             'Text': parse.Text,
             'QuotedString': QuotedString,
             }
tt_grammar = resource_string(__name__, 'grammar_tt.txt').decode('UTF-8')
tt_parser = parsley.makeGrammar(tt_grammar, parser_ns)


def parse_tt(source):
    try:
        return parse.add_positions(list(tt_parser(source).template()))
    except parsley.ParseError as e:
        raise PigletParseError(e.formatError())


def make_tree(parse_result):
    """
    From ``parse_result``, a list of parsed tokens create a tree of
    elements
    """
    stack = [Container()]
    head = stack[-1]
    for item in parse_result:
        if isinstance(item, Statement) and item.is_open_tag():
            head.children.append(item)
            stack.append(item)
            head = item

        elif isinstance(item, Statement) and item.is_close_tag():
            # else statements don't have associated end tags - the end tag
            # should cause the parent if statement to close instead
            if head.name == 'else':
                head = stack.pop()
            name = item.name[3:]
            if name and name != head.name:
                raise parse.PigletParseError("Unexpected {} at {}:{}"
                                             .format(item.source,
                                                     item.pos.line,
                                                     item.pos.char))
            head.end_tag = item
            head.end_tag_pos = item.pos
            stack.pop()
            head = stack[-1]
        else:
            head.children.append(item)
    if len(stack) > 1:
        raise parse.PigletParseError("Missing end tag for {}, "
                                     "opened at {}:{}"
                                     .format(stack[-1].source,
                                             stack[-1].pos.line,
                                             stack[-1].pos.char))
    return stack[0]


def compile_intermediate(parse_result):
    tree = make_tree(parse_result)
    strip_ws_on_directive_lines(tree)
    cx.strip_spaces_between_directives(tree,
                                       element_cls=Statement,
                                       is_directive=lambda x: isinstance(x, Statement))
    compiled = _compile_node(tree)
    cx._strip_unwanted_nodes(compiled)
    cx._concatentate_adjacent_text_nodes(compiled)
    cx._postprocess(compiled)
    return compiled


def strip_ws_on_directive_lines(node):

    # Remove whitespace until after first newline
    trim_line_start = partial(re.compile(r'\A[ \t]*(\r\n|\n|\r)').sub, r'')

    # Remove whitespace after last newline
    trim_line_end = partial(re.compile(r'([\r\n])[ \t]*\Z').sub, r'\1')

    # Remove leading whitespace
    trim_all_start = partial(re.compile(r'\A\s*', re.S).sub, r'')

    # Remove trailing whitespace
    trim_all_end = partial(re.compile(r'\s*\Z', re.S).sub, r'')

    trim_nothing = lambda s: s

    trim_start = {'+': trim_nothing,
                  '-': trim_all_start,
                   None: trim_line_start}
    trim_end = {'+': trim_nothing,
                '-': trim_all_end,
                None: trim_line_end}

    if isinstance(node, Statement):
        if node.children:
            # Remove whitespace until after first newline inside a statement
            c = node.children[0]
            if isinstance(c, parse.Text):
                c.content = trim_start[node.ws_after](c.content)

            # Remove whitespace after the last newline inside a statement
            c = node.children[-1]
            if isinstance(c, parse.Text):
                c.content = trim_end[node.end_tag.ws_before](c.content)

    for p, n in zip(node.children, node.children[1:]):

        # Remove whitespace preceding a statement
        if isinstance(p, parse.Text) and isinstance(n, Statement):
            p.content = trim_end[n.ws_before](p.content)

        # Remove whitespace following a statement
        if isinstance(p, Statement) and isinstance(n, parse.Text):
            n.content = trim_start[p.end_tag.ws_after](n.content)

    for item in node.children:
        if hasattr(item, 'children'):
            strip_ws_on_directive_lines(item)


def compile_extends(stmt, href):
    return im.ExtendsNode(href=href,
                          pos=stmt.pos,
                          children=[_compile_node(c) for c in stmt.children])


def compile_block(stmt, name):
    return im.BlockNode(name=name,
                        pos=stmt.pos,
                        children=[_compile_node(c) for c in stmt.children])


def compile_for(stmt, *args):
    return im.ForNode(each=' '.join(args).strip(),
                      pos=stmt.pos)


def compile_if(stmt, *args):
    n = im.IfNode(test=' '.join(args).strip(),
                  pos=stmt.pos)

    children = stmt.children
    if children:
        last = children[-1]
        if isinstance(last, Statement) and last.name == 'else':
            children = children[:-1]
            n.else_ = im.ElseNode(
                pos=last.pos,
                children=[_compile_node(c) for c in last.children])
    n.children = [_compile_node(c) for c in children]
    return n


def compile_with(stmt, *args):
    return im.WithNode(vars=stmt.raw_args,
                       children=[_compile_node(c) for c in stmt.children],
                       pos=stmt.pos)


def compile_trans(stmt, message=None):
    return im.TranslationNode(
        message=message,
        children=[_compile_node(c) for c in stmt.children],
        pos=stmt.pos)


def compile_transname(stmt, name):
    return im.TranslationPlaceholder(
        name=name,
        children=[_compile_node(c) for c in stmt.children],
        pos=stmt.pos)


DIRECTIVES = {
    'extends': compile_extends,
    'block': compile_block,
    'for': compile_for,
    'if': compile_if,
    'with': compile_with,
    'trans': compile_trans,
    'transname': compile_transname,
}


def split_args(args):
    for a in args:
        if isinstance(a, QuotedString):
            yield a.content
        else:
            for b in a.strip().split():
                if b:
                    yield b


def _compile_node(node):
    """
    Compile a single node to intermediate representation
    """

    if isinstance(node, Container):
        root = im.ContainerNode()
        root.extend(map(_compile_node, node.children))
        return root

    elif isinstance(node, parse.Text):
        return cx._compile_text(node)

    elif isinstance(node, Statement):
        args = split_args(node.args)
        return DIRECTIVES[node.name](node, *args)

    else:
        raise NotImplementedError(node)
